from datetime import datetime, timedelta
from numbers import Number
from typing import Union, Mapping, Type, Dict, Callable, Tuple, List

KustoTypes = Union[str, Number, bool, datetime, Mapping, List, Tuple, timedelta]
# TODO: Unhandled data types: guid, decimal


class TypeRegistrar:
    """
    A factory for annotations that are used to create a mapping between Kusto types and python types and functions.
    Each annotation must be called with a Kusto type as a parameter. The `for_obj` and `for_type` methods
    can then be used to retrieve the python type or function corresponding to a given Kusto type.
    """
    registry: Dict[Type[KustoTypes], Callable]

    def __init__(self, name: str) -> None:
        """
        :param name: Name is used for better logging and clearer errors
        """
        self.name = name
        self.registry = {}

    def __repr__(self) -> str:
        return self.name

    def __call__(self, *types: Type[KustoTypes]) -> Callable:
        def inner(wrapped):
            for t in types:
                previous = self.registry.setdefault(t, wrapped)
                if previous is not wrapped:
                    raise TypeError("{}: type already registered: {}".format(self, t.__name__))
            return wrapped

        return inner

    def for_obj(self, obj: KustoTypes) -> Callable:
        """
        Given an object of Kusto type, retrieve the python type or function associated with the object's type, and call
        it with the given object as a parameter

        :param obj: An object of Kusto type
        :return: Associated python object
        """
        for registered_type, registered_callable in self.registry.items():
            if isinstance(obj, registered_type):
                return registered_callable(obj)
        raise ValueError("{}: no registered callable for object {} of type {}".format(self, obj, type(obj).__name__))

    def for_type(self, t: Type[KustoTypes]) -> Callable:
        """
        Given a Kusto type, retrieve the associated python type or function

        :param t: A Kusto type
        :return: Associated python object
        """
        for registered_type, registered_callable in self.registry.items():
            if issubclass(t, registered_type):
                return registered_callable
        raise ValueError("{}: no registered callable for type {}".format(self, t.__name__))


kql_converter = TypeRegistrar("KQL Converter")
plain_expression = TypeRegistrar("Plain expression")
aggregation_expression = TypeRegistrar("Aggregation expression")



from typing import Union, List, Tuple
# noinspection PyProtectedMember
from urllib.parse import urlparse

# noinspection PyProtectedMember
from azure.kusto.data._response import KustoResponseDataSet
from azure.kusto.data.request import KustoClient, KustoConnectionStringBuilder, ClientRequestProperties

from pykusto.kql_converters import KQL


class PyKustoClient:
    """
    Handle to a Kusto cluster
    """
    _client: KustoClient
    _cluster_name: str

    def __init__(self, client_or_cluster: Union[str, KustoClient]) -> None:
        """
        Create a new handle to Kusto cluster

        :param client_or_cluster: Either a KustoClient object, or a cluster name. In case a cluster name is given,
            a KustoClient is generated with AAD device authentication
        """
        if isinstance(client_or_cluster, KustoClient):
            self._client = client_or_cluster
            # noinspection PyProtectedMember
            self._cluster_name = urlparse(client_or_cluster._query_endpoint).netloc  # TODO neater way
        else:
            self._client = self._get_client_for_cluster(client_or_cluster)
            self._cluster_name = client_or_cluster

    def execute(self, database: str, query: KQL, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
        return self._client.execute(database, query, properties)

    def show_databases(self) -> Tuple[str, ...]:
        res: KustoResponseDataSet = self.execute('', KQL('.show databases'))
        return tuple(r[0] for r in res.primary_results[0].rows)

    def __getitem__(self, database_name: str) -> 'Database':
        return Database(self, database_name)

    def get_cluster_name(self) -> str:
        return self._cluster_name

    @staticmethod
    def _get_client_for_cluster(cluster: str) -> KustoClient:
        return KustoClient(KustoConnectionStringBuilder.with_aad_device_authentication(cluster))


class Database:
    """
    Handle to a Kusto database
    """
    client: PyKustoClient
    name: str

    def __init__(self, client: PyKustoClient, name: str) -> None:
        self.client = client
        self.name = name

    def execute(self, query: KQL, properties: ClientRequestProperties = None) -> KustoResponseDataSet:
        return self.client.execute(self.name, query, properties)

    def show_tables(self) -> Tuple[str, ...]:
        res: KustoResponseDataSet = self.execute(KQL('.show tables'))
        return tuple(r[0] for r in res.primary_results[0].rows)

    def get_tables(self, *tables: str):
        return Table(self, tables)

    def __getitem__(self, table_name: str) -> 'Table':
        return self.get_tables(table_name)


class Table:
    """
    Handle to a Kusto table
    """
    database: Database
    tables: Union[str, List[str], Tuple[str, ...]]

    def __init__(self, database: Database, tables: Union[str, List[str], Tuple[str, ...]]) -> None:
        """
        Create a new handle to a Kusto table

        :param database: Database object
        :param tables: Either a single table name, or a list of tables. If more than one table is given OR the table
            name contains a wildcard, the Kusto 'union' statement will be used.
        """

        self.database = database
        self.tables = [tables] if isinstance(tables, str) else tables

    def get_table(self) -> KQL:
        result = KQL(', '.join(self.tables))
        if '*' in result or ',' in result:
            result = KQL('union ' + result)
        return result

    def get_full_table(self) -> KQL:
        assert len(self.tables) > 0
        if len(self.tables) == 1 and not any('*' in t for t in self.tables):
            return self._format_full_table_name(self.tables[0])
        else:
            return KQL("union " + ", ".join(self._format_full_table_name(t) for t in self.tables))

    def _format_full_table_name(self, table):
        table_format_str = 'cluster("{}").database("{}").table("{}")'
        return KQL(
            table_format_str.format(self.database.client.get_cluster_name(), self.database.name, table))

    def execute(self, rendered_query: KQL) -> KustoResponseDataSet:
        return self.database.execute(rendered_query)

    def show_columns(self) -> Tuple[Tuple[str, str], ...]:
        res: KustoResponseDataSet = self.execute(KQL('.show table {}'.format(self.get_table())))
        return tuple(
            (
                r[0],  # Column name
                r[1],  # Column type
            )
            for r in res.primary_results[0].rows
        )



from datetime import datetime, timedelta
from numbers import Number
from typing import Any, List, Tuple, Mapping, Optional, Type
from typing import Union

from pykusto.kql_converters import KQL
from pykusto.type_utils import plain_expression, aggregation_expression, KustoTypes, kql_converter

ExpressionType = Union[KustoTypes, 'BaseExpression']
StringType = Union[str, 'StringExpression']
BooleanType = Union[bool, 'BooleanExpression']
NumberType = Union[int, float, 'NumberExpression']
ArrayType = Union[List, Tuple, 'ArrayExpression']
MappingType = Union[Mapping, 'MappingExpression']
DatetimeType = Union[datetime, 'DatetimeExpression']
TimespanType = Union[timedelta, 'TimespanExpression']
DynamicType = Union[ArrayType, MappingType]
OrderType = Union[DatetimeType, TimespanType, NumberType, StringType]


# All classes in the same file to prevent circular dependencies

def _subexpr_to_kql(obj: ExpressionType) -> KQL:
    """
    Convert the given expression to KQL, enclosing it in parentheses if it is a compound expression. This guarantees
    correct evaluation order. When parentheses are not needed, for example when the expressions is used as an argument
    to a function, use `to_kql` instead.

    :param obj: Expression to convert to KQL
    :return: KQL that represents the given expression
    """
    if isinstance(obj, BaseExpression):
        return obj.as_subexpression()
    return to_kql(obj)


class BaseExpression:
    kql: KQL

    def __new__(cls, *args, **kwargs):
        if cls is 'BaseExpression':
            raise TypeError("BaseExpression is abstract")
        return object.__new__(cls)

    def __init__(self, kql: Union[KQL, 'BaseExpression']) -> None:
        if isinstance(kql, BaseExpression):
            self.kql = kql.kql
        elif not isinstance(kql, str):
            raise ValueError("Either expression or KQL required")
        self.kql = kql

    def __repr__(self) -> str:
        return str(self.kql)

    def as_subexpression(self) -> KQL:
        return KQL('({})'.format(self.kql))

    def gettype(self) -> 'StringExpression':
        return StringExpression(KQL('gettype({})'.format(self.kql)))

    def __hash__(self) -> 'StringExpression':
        return StringExpression(KQL('hash({})'.format(self.kql)))

    def hash_sha256(self) -> 'StringExpression':
        return StringExpression(KQL('hash_sha256({})'.format(self.kql)))

    def is_empty(self) -> 'BooleanExpression':
        return BooleanExpression(KQL('isempty({})'.format(self.kql)))

    def is_not_empty(self) -> 'BooleanExpression':
        return BooleanExpression(KQL('isnotempty({})'.format(self.kql)))

    def has(self, exp: str) -> 'BooleanExpression':
        # The pattern for the search expression must be a constant string.
        return BooleanExpression(KQL('{} has \"{}\"'.format(self.kql, exp)))

    @staticmethod
    def base_binary_op(
            left: ExpressionType, operator: str, right: ExpressionType, result_type: Type[KustoTypes]
    ) -> 'BaseExpression':
        registrar = plain_expression
        if isinstance(left, AggregationExpression) or isinstance(right, AggregationExpression):
            registrar = aggregation_expression
        return registrar.for_type(result_type)(KQL('{}{}{}'.format(
            _subexpr_to_kql(left), operator, _subexpr_to_kql(right))
        ))

    def __eq__(self, other: ExpressionType) -> 'BooleanExpression':
        return BooleanExpression.binary_op(self, ' == ', other)

    def __ne__(self, other: ExpressionType) -> 'BooleanExpression':
        return BooleanExpression.binary_op(self, ' != ', other)

    def is_in(self, other: ArrayType) -> 'BooleanExpression':
        return BooleanExpression.binary_op(self, ' in ', other)

    def is_null(self) -> 'BooleanExpression':
        return BooleanExpression(KQL('isnull({})'.format(self.kql)))

    def is_not_null(self) -> 'BooleanExpression':
        return BooleanExpression(KQL('isnotnull({})'.format(self.kql)))

    def __contains__(self, other: Any) -> bool:
        """
        Deliberately not implemented, because "not in" inverses the result of this method, and there is no way to
        override it
        """
        raise NotImplementedError("Instead use 'is_in' or 'contains'")

    def to_bool(self) -> 'BooleanExpression':
        return BooleanExpression(KQL('tobool({})'.format(self.kql)))

    def to_string(self) -> 'StringExpression':
        return StringExpression(KQL('tostring({})'.format(self.kql)))

    def assign_to(self, *columns: 'Column') -> 'AssignmentBase':
        if len(columns) == 0:
            # Unspecified column name
            return AssignmentBase(None, self)
        if len(columns) == 1:
            return AssignmentToSingleColumn(columns[0], self)
        raise ValueError("Only arrays can be assigned to multiple columns")


@plain_expression(bool)
class BooleanExpression(BaseExpression):
    @staticmethod
    def binary_op(left: ExpressionType, operator: str, right: ExpressionType) -> 'BooleanExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, bool)

    def __and__(self, other: BooleanType) -> 'BooleanExpression':
        return BooleanExpression.binary_op(self, ' and ', other)

    def __or__(self, other: BooleanType) -> 'BooleanExpression':
        return BooleanExpression.binary_op(self, ' or ', other)

    def __invert__(self) -> 'BooleanExpression':
        return BooleanExpression(KQL('not({})'.format(self.kql)))


@plain_expression(Number)
class NumberExpression(BaseExpression):
    @staticmethod
    def binary_op(left: NumberType, operator: str, right: NumberType) -> 'NumberExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, Number)

    def __lt__(self, other: NumberType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' < ', other)

    def __le__(self, other: NumberType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' <= ', other)

    def __gt__(self, other: NumberType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' > ', other)

    def __ge__(self, other: NumberType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' >= ', other)

    def __add__(self, other: NumberType) -> 'NumberExpression':
        return NumberExpression.binary_op(self, ' + ', other)

    def __sub__(self, other: NumberType) -> 'NumberExpression':
        return NumberExpression.binary_op(self, ' - ', other)

    def __mul__(self, other: NumberType) -> 'NumberExpression':
        return NumberExpression.binary_op(self, ' * ', other)

    def __truediv__(self, other: NumberType) -> 'NumberExpression':
        return NumberExpression.binary_op(self, ' / ', other)

    def __mod__(self, other: NumberType) -> 'NumberExpression':
        return NumberExpression.binary_op(self, ' % ', other)

    def __neg__(self) -> 'NumberExpression':
        return NumberExpression(KQL('-{}'.format(self.kql)))

    def __abs__(self) -> 'NumberExpression':
        return NumberExpression(KQL('abs({})'.format(self.kql)))

    def between(self, lower: NumberType, upper: NumberType) -> BooleanExpression:
        return BooleanExpression(KQL('{} between ({} .. {})'.format(
            self.kql, _subexpr_to_kql(lower), _subexpr_to_kql(upper)
        )))

    def acos(self) -> 'NumberExpression':
        return NumberExpression(KQL('acos({})'.format(self.kql)))

    def cos(self) -> 'NumberExpression':
        return NumberExpression(KQL('cos({})'.format(self.kql)))

    def floor(self, round_to: NumberType) -> 'NumberExpression':
        return NumberExpression(KQL('floor({}, {})'.format(self.kql, _subexpr_to_kql(round_to))))

    def bin(self, round_to: NumberType) -> 'BaseExpression':
        return NumberExpression(KQL('bin({}, {})'.format(self.kql, _subexpr_to_kql(round_to))))

    def bin_at(self, round_to: NumberType, fixed_point: NumberType) -> 'BaseExpression':
        return NumberExpression(KQL('bin_at({}, {}, {})'.format(self.kql,
                                                                _subexpr_to_kql(round_to),
                                                                _subexpr_to_kql(fixed_point))))

    def bin_auto(self) -> 'BaseExpression':
        return NumberExpression(KQL('bin_auto({})'.format(self.kql)))

    def ceiling(self) -> 'NumberExpression':
        return NumberExpression(KQL('ceiling({})'.format(self.kql)))

    def exp(self) -> 'NumberExpression':
        return NumberExpression(KQL('exp({})'.format(self.kql)))

    def exp10(self) -> 'NumberExpression':
        return NumberExpression(KQL('exp10({})'.format(self.kql)))

    def exp2(self) -> 'NumberExpression':
        return NumberExpression(KQL('exp2({})'.format(self.kql)))

    def isfinite(self) -> BooleanExpression:
        return BooleanExpression(KQL('isfinite({})'.format(self.kql)))

    def isinf(self) -> BooleanExpression:
        return BooleanExpression(KQL('isinf({})'.format(self.kql)))

    def isnan(self) -> BooleanExpression:
        return BooleanExpression(KQL('isnan({})'.format(self.kql)))

    def log(self) -> 'NumberExpression':
        return NumberExpression(KQL('log({})'.format(self)))

    def log10(self) -> 'NumberExpression':
        return NumberExpression(KQL('log10({})'.format(self)))

    def log2(self) -> 'NumberExpression':
        return NumberExpression(KQL('log2({})'.format(self)))

    def loggamma(self) -> 'NumberExpression':
        return NumberExpression(KQL('loggamma({})'.format(self)))

    def round(self, precision: NumberType = None) -> 'NumberExpression':
        return NumberExpression(KQL(
            ('round({}, {})' if precision is None else 'round({}, {})').format(self.kql, to_kql(precision))
        ))


@plain_expression(str)
class StringExpression(BaseExpression):
    @staticmethod
    def binary_op(left: ExpressionType, operator: str, right: ExpressionType) -> 'StringExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, str)

    def __len__(self) -> NumberExpression:
        return self.string_size()

    def string_size(self) -> NumberExpression:
        return NumberExpression(KQL('string_size({})'.format(self.kql)))

    def is_empty(self) -> BooleanExpression:
        return BooleanExpression(KQL('isempty({})'.format(self.kql)))

    def __add__(self, other: StringType) -> 'StringExpression':
        return StringExpression.binary_op(self, ' + ', other)

    @staticmethod
    def concat(*strings: StringType) -> 'StringExpression':
        return StringExpression(KQL('strcat({})'.format(', '.join('{}'.format(
            _subexpr_to_kql(s)
        ) for s in strings))))

    def split(self, delimiter: StringType, requested_index: NumberType = None) -> 'ArrayExpression':
        if requested_index is None:
            return ArrayExpression(KQL('split({}, {})'.format(to_kql(self.kql), to_kql(delimiter))))
        return ArrayExpression(KQL('split({}, {}, {})'.format(
            to_kql(self.kql), to_kql(delimiter), to_kql(requested_index)
        )))

    def equals(self, other: StringType, case_sensitive: bool = False) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' == ' if case_sensitive else ' =~ ', other)

    def not_equals(self, other: StringType, case_sensitive: bool = False) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' != ' if case_sensitive else ' !~ ', other)

    def matches(self, regex: StringType) -> 'BooleanExpression':
        return BooleanExpression.binary_op(self, ' matches regex ', regex)

    def contains(self, other: StringType, case_sensitive: bool = False) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' contains_cs ' if case_sensitive else ' contains ', other)

    def not_contains(self, other: StringType, case_sensitive: bool = False) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' !contains_cs ' if case_sensitive else ' !contains ', other)

    def startswith(self, other: StringType, case_sensitive: bool = False) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' startswith_cs ' if case_sensitive else ' startswith ', other)

    def endswith(self, other: StringType, case_sensitive: bool = False) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' endswith_cs ' if case_sensitive else ' endswith ', other)

    def to_int(self) -> NumberExpression:
        return NumberExpression(KQL('toint({})'.format(self.kql)))

    def to_long(self) -> NumberExpression:
        return NumberExpression(KQL('tolong({})'.format(self.kql)))

    def lower(self) -> 'StringExpression':
        return StringExpression(KQL('tolower({})'.format(self.kql)))

    def upper(self) -> 'StringExpression':
        return StringExpression(KQL('toupper({})'.format(self.kql)))

    def is_utf8(self) -> BooleanExpression:
        return BooleanExpression(KQL('isutf8({})'.format(self.kql)))


@plain_expression(datetime)
class DatetimeExpression(BaseExpression):
    @staticmethod
    def binary_op(left: ExpressionType, operator: str, right: ExpressionType) -> 'DatetimeExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, datetime)

    def __lt__(self, other: DatetimeType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' < ', other)

    def __le__(self, other: DatetimeType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' <= ', other)

    def __gt__(self, other: DatetimeType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' > ', other)

    def __ge__(self, other: DatetimeType) -> BooleanExpression:
        return BooleanExpression.binary_op(self, ' >= ', other)

    def __add__(self, other: TimespanType) -> 'DatetimeExpression':
        return DatetimeExpression.binary_op(self, ' + ', other)

    def __sub__(self, other: Any) -> BaseExpression:
        raise NotImplementedError("Instead use 'date_diff' or 'subtract_timespan'")

    def date_diff(self, other: DatetimeType) -> 'TimespanExpression':
        return TimespanExpression.binary_op(self, ' - ', other)

    def subtract_timespan(self, other: TimespanType) -> 'DatetimeExpression':
        return DatetimeExpression.binary_op(self, ' - ', other)

    def between(self, lower: DatetimeType, upper: DatetimeType) -> BooleanExpression:
        return BooleanExpression(KQL('{} between ({} .. {})'.format(
            self.kql, _subexpr_to_kql(lower), _subexpr_to_kql(upper)
        )))

    def floor(self, round_to: TimespanType) -> 'DatetimeExpression':
        return DatetimeExpression(KQL('floor({}, {})'.format(self.kql, _subexpr_to_kql(round_to))))

    def bin(self, round_to: TimespanType) -> 'BaseExpression':
        return DatetimeExpression(KQL('bin({}, {})'.format(self.kql, _subexpr_to_kql(round_to))))

    def bin_at(self, round_to: TimespanType, fixed_point: DatetimeType) -> 'BaseExpression':
        return DatetimeExpression(KQL('bin_at({}, {}, {})'.format(
            self.kql, _subexpr_to_kql(round_to), to_kql(fixed_point)
        )))

    def bin_auto(self) -> 'BaseExpression':
        return DatetimeExpression(KQL('bin_auto({})'.format(self.kql)))

    def endofday(self, offset: NumberType = None) -> 'DatetimeExpression':
        if offset is None:
            res = 'endofday({})'.format(self.kql)
        else:
            res = 'endofday({}, {})'.format(self.kql, _subexpr_to_kql(offset))
        return DatetimeExpression(KQL(res))

    def endofmonth(self, offset: NumberType = None) -> 'DatetimeExpression':
        if offset is None:
            res = 'endofmonth({})'.format(self.kql)
        else:
            res = 'endofmonth({}, {})'.format(self.kql, _subexpr_to_kql(offset))
        return DatetimeExpression(KQL(res))

    def endofweek(self, offset: NumberType = None) -> 'DatetimeExpression':
        if offset is None:
            res = 'endofweek({})'.format(self.kql)
        else:
            res = 'endofweek({}, {})'.format(self.kql, _subexpr_to_kql(offset))
        return DatetimeExpression(KQL(res))

    def endofyear(self, offset: NumberType = None) -> 'DatetimeExpression':
        if offset is None:
            res = 'endofyear({})'.format(self.kql)
        else:
            res = 'endofyear({}, {})'.format(self.kql, _subexpr_to_kql(offset))
        return DatetimeExpression(KQL(res))

    def format_datetime(self, format_string: StringType) -> StringExpression:
        return StringExpression(KQL('format_datetime({}, {})'.format(self.kql, _subexpr_to_kql(format_string))))

    def getmonth(self) -> NumberExpression:
        return NumberExpression(KQL('getmonth({})'.format(self.kql)))

    def getyear(self) -> NumberExpression:
        return NumberExpression(KQL('getyear({})'.format(self.kql)))

    def hourofday(self) -> NumberExpression:
        return NumberExpression(KQL('hourofday({})'.format(self.kql)))

    def startofday(self, offset: NumberType = None) -> 'DatetimeExpression':
        return DatetimeExpression(KQL(
            ('startofday({})' if offset is None else 'startofday({}, {})').format(self.kql, to_kql(offset))
        ))

    def startofmonth(self, offset: NumberType = None) -> 'DatetimeExpression':
        return DatetimeExpression(KQL(
            ('startofmonth({})' if offset is None else 'startofmonth({}, {})').format(self.kql, to_kql(offset))
        ))

    def startofweek(self, offset: NumberType = None) -> 'DatetimeExpression':
        return DatetimeExpression(KQL(
            ('startofweek({})' if offset is None else 'startofweek({}, {})').format(self.kql, to_kql(offset))
        ))

    def startofyear(self, offset: NumberType = None) -> 'DatetimeExpression':
        return DatetimeExpression(KQL(
            ('startofyear({})' if offset is None else 'startofyear({}, {})').format(self.kql, to_kql(offset))
        ))


@plain_expression(timedelta)
class TimespanExpression(BaseExpression):
    @staticmethod
    def binary_op(left: ExpressionType, operator: str, right: ExpressionType) -> 'TimespanExpression':
        # noinspection PyTypeChecker
        return BaseExpression.base_binary_op(left, operator, right, timedelta)

    def __add__(self, other: TimespanType) -> 'TimespanExpression':
        return TimespanExpression.binary_op(self, ' + ', other)

    def __sub__(self, other: TimespanType) -> 'TimespanExpression':
        return TimespanExpression.binary_op(self, ' - ', other)

    def ago(self) -> DatetimeExpression:
        return DatetimeExpression(KQL('ago({})'.format(_subexpr_to_kql(self))))

    def bin(self, round_to: TimespanType) -> 'BaseExpression':
        return TimespanExpression(KQL('bin({}, {})'.format(self.kql, _subexpr_to_kql(round_to))))

    def bin_at(self, round_to: TimespanType, fixed_point: TimespanType) -> 'BaseExpression':
        return TimespanExpression(KQL('bin_at({}, {}, {})'.format(
            self.kql, to_kql(round_to), to_kql(fixed_point)
        )))

    def bin_auto(self) -> 'BaseExpression':
        return TimespanExpression(KQL('bin_auto({})'.format(self.kql)))

    def format_timespan(self, format_string: StringType) -> StringExpression:
        return StringExpression(KQL('format_timespan({}, {})'.format(self.kql, to_kql(format_string))))

    def between(self, lower: TimespanType, upper: TimespanType) -> BooleanExpression:
        return BooleanExpression(KQL('{} between ({} .. {})'.format(
            self.kql, _subexpr_to_kql(lower), _subexpr_to_kql(upper)
        )))


@plain_expression(List, Tuple)
class ArrayExpression(BaseExpression):
    def __getitem__(self, index: NumberType) -> 'AnyExpression':
        return AnyExpression(KQL('{}[{}]'.format(self.kql, _subexpr_to_kql(index))))

    def __len__(self) -> NumberExpression:
        return self.array_length()

    def array_length(self) -> NumberExpression:
        return NumberExpression(KQL('array_length({})'.format(self.kql)))

    def contains(self, other: ExpressionType) -> 'BooleanExpression':
        return BooleanExpression.binary_op(other, ' in ', self)

    @staticmethod
    def pack_array(*elements: ExpressionType) -> 'ArrayExpression':
        return ArrayExpression(KQL('pack_array({})'.format(
            ', '.join('{}'.format(_subexpr_to_kql(e) for e in elements))
        )))

    def __getitem__(self, index: NumberType) -> BaseExpression:
        return AnyExpression(KQL('{}[{}]'.format(self.kql, _subexpr_to_kql(index))))

    def assign_to(self, *columns: 'Column') -> 'AssignmentBase':
        if len(columns) <= 1:
            return super().assign_to(*columns)
        return AssignmentToMultipleColumns(columns, self)


@plain_expression(Mapping)
class MappingExpression(BaseExpression):
    def __getitem__(self, index: StringType) -> 'AnyExpression':
        return AnyExpression(KQL('{}[{}]'.format(self.kql, _subexpr_to_kql(index))))

    def __getattr__(self, name: str) -> 'AnyExpression':
        return AnyExpression(KQL('{}.{}'.format(self.kql, name)))

    def keys(self) -> ArrayExpression:
        return ArrayExpression(KQL('bag_keys({})'.format(self.kql)))

    @staticmethod
    def pack(**kwargs: ExpressionType) -> 'MappingExpression':
        return MappingExpression(KQL('pack({})'.format(
            ', '.join('"{}", {}'.format(k, _subexpr_to_kql(v)) for k, v in kwargs)
        )))


class DynamicExpression(ArrayExpression, MappingExpression):
    def __getitem__(self, index: Union[StringType, NumberType]) -> 'AnyExpression':
        return AnyExpression(KQL('{}[{}]'.format(self.kql, _subexpr_to_kql(index))))


class AnyExpression(
    NumberExpression, BooleanExpression,
    StringExpression, DynamicExpression,
    DatetimeExpression, TimespanExpression
):
    pass


class AggregationExpression(BaseExpression):
    def __new__(cls, *args, **kwargs):
        if cls is 'AggregationExpression':
            raise TypeError("AggregationExpression is abstract")
        return object.__new__(cls)

    def assign_to(self, *columns: 'Column') -> 'AssignmentFromAggregationToColumn':
        if len(columns) == 0:
            # Unspecified column name
            return AssignmentFromAggregationToColumn(None, self)
        if len(columns) == 1:
            return AssignmentFromAggregationToColumn(columns[0], self)
        raise ValueError("Aggregations cannot be assigned to multiple columns")

    def as_subexpression(self) -> KQL:
        return self.kql


@aggregation_expression(bool)
class BooleanAggregationExpression(AggregationExpression, BooleanExpression):
    pass


@aggregation_expression(Number)
class NumberAggregationExpression(AggregationExpression, NumberExpression):
    pass


@aggregation_expression(str)
class StringAggregationExpression(AggregationExpression, StringExpression):
    pass


@aggregation_expression(datetime)
class DatetimeAggregationExpression(AggregationExpression, DatetimeExpression):
    pass


@aggregation_expression(timedelta)
class TimespanAggregationExpression(AggregationExpression, TimespanExpression):
    pass


@aggregation_expression(List, Tuple)
class ArrayAggregationExpression(AggregationExpression, ArrayExpression):
    pass


@aggregation_expression(Mapping)
class MappingAggregationExpression(AggregationExpression, MappingExpression):
    pass


class AssignmentBase:
    _lvalue: Optional[KQL]
    _rvalue: KQL

    def __init__(self, lvalue: Optional[KQL], rvalue: ExpressionType) -> None:
        self._lvalue = lvalue
        self._rvalue = rvalue.as_subexpression()

    def to_kql(self) -> KQL:
        if self._lvalue is None:
            # Unspecified column name
            return self._rvalue
        return KQL('{} = {}'.format(self._lvalue, self._rvalue))

    @staticmethod
    def assign(expression: ExpressionType, *columns: 'Column') -> 'AssignmentBase':
        if len(columns) == 0:
            raise ValueError("Provide at least one column")
        if len(columns) == 1:
            return AssignmentToSingleColumn(columns[0], expression)
        return AssignmentToMultipleColumns(columns, expression)


class AssignmentToSingleColumn(AssignmentBase):
    def __init__(self, column: 'Column', expression: ExpressionType) -> None:
        super().__init__(column.kql, expression)


class AssignmentFromColumnToColumn(AssignmentToSingleColumn):
    def __init__(self, target: 'Column', source: 'Column') -> None:
        super().__init__(target, source)


class AssignmentToMultipleColumns(AssignmentBase):
    def __init__(self, columns: Union[List['Column'], Tuple['Column']], expression: ArrayType) -> None:
        super().__init__(KQL('({})'.format(', '.join(c.kql for c in columns))), expression)


class AssignmentFromAggregationToColumn(AssignmentBase):
    def __init__(self, column: Optional['Column'], aggregation: AggregationExpression) -> None:
        super().__init__(None if column is None else column.kql, aggregation)


class Column(AnyExpression):
    _name: str

    def __init__(self, name: str) -> None:
        super().__init__(KQL("['{}']".format(name) if '.' in name else name))
        self._name = name

    def as_subexpression(self) -> KQL:
        return self.kql

    def __len__(self) -> NumberExpression:
        raise NotImplementedError("Column type unknown, instead use 'string_size' or 'array_length'")

    def assign_to(self, *columns: 'Column') -> 'AssignmentBase':
        if len(columns) == 0:
            return super().assign_to()
        if len(columns) == 1:
            return AssignmentFromColumnToColumn(columns[0], self)
        return ArrayExpression.assign_to(self, *columns)


class ColumnGenerator:
    def __getattr__(self, name: str) -> Column:
        return Column(name)

    def __getitem__(self, name: str) -> Column:
        return Column(name)


# Recommended usage: from pykusto.expressions import column_generator as col
# TODO: Is there a way to enforce this to be a singleton?
column_generator = ColumnGenerator()


def to_kql(obj: ExpressionType) -> KQL:
    """
    Convert the given expression to KQL. If this is a subexpression of a greater expression, neighboring operators might
    take precedence over operators included in this expression, causing an incorrect evaluation order.
    If this is a concern, use `_subexpr_to_kql` instead, which will enclose this expression in parentheses if it is
    a compound expression.

    :param obj: Expression to convert to KQL
    :return: KQL that represents the given expression
    """
    if isinstance(obj, BaseExpression):
        return obj.kql
    return kql_converter.for_obj(obj)



from abc import abstractmethod
from copy import copy, deepcopy
from enum import Enum
from itertools import chain
from types import FunctionType
from typing import Tuple, List, Union, Optional

from azure.kusto.data.helpers import dataframe_from_result_table

from pykusto.client import Table
from pykusto.expressions import BooleanType, ExpressionType, AggregationExpression, OrderType, \
    StringType, AssignmentBase, AssignmentFromAggregationToColumn, AssignmentToSingleColumn, Column, BaseExpression, \
    AssignmentFromColumnToColumn, AnyExpression, to_kql
from pykusto.kql_converters import KQL
from pykusto.logger import logger
from pykusto.udf import stringify_python_func


class Order(Enum):
    ASC = "asc"
    DESC = "desc"


class Nulls(Enum):
    FIRST = "first"
    LAST = "last"


class JoinKind(Enum):
    INNERUNIQUE = "innerunique"
    INNER = "inner"
    LEFTOUTER = "leftouter"
    RIGHTOUTER = "rightouter"
    FULLOUTER = "fullouter"
    LEFTANTI = "leftanti"
    ANTI = "anti"
    LEFTANTISEMI = "leftantisemi"
    RIGHTANTI = "rightanti"
    RIGHTANTISEMI = "rightantisemi"
    LEFTSEMI = "leftsemi"
    RIGHTSEMI = "rightsemi"


class BagExpansion(Enum):
    BAG = "bag"
    ARRAY = "array"


class Query:
    _head: Optional['Query']
    _table: Optional[Table]

    def __init__(self, head=None) -> None:
        self._head = head if isinstance(head, Query) else None
        self._table = head if isinstance(head, Table) else None

    def __add__(self, other: 'Query'):
        self_copy = deepcopy(self)
        other_copy = deepcopy(other)

        other_base = other_copy
        while other_base._head is not None:
            if other_base._head._head is None:
                break
            other_base = other_base._head
        other_base._head = self_copy
        return other_copy

    def __deepcopy__(self, memo):
        new_object = copy(self)
        if self._head is not None:
            new_object._head = self._head.__deepcopy__(memo)
        return new_object

    def where(self, predicate: BooleanType) -> 'WhereQuery':
        return WhereQuery(self, predicate)

    def take(self, num_rows: int) -> 'TakeQuery':
        return TakeQuery(self, num_rows)

    def limit(self, num_rows: int) -> 'LimitQuery':
        return LimitQuery(self, num_rows)

    def sample(self, num_rows: int) -> 'SampleQuery':
        return SampleQuery(self, num_rows)

    def count(self) -> 'CountQuery':
        return CountQuery(self)

    def sort_by(self, col: OrderType, order: Order = None, nulls: Nulls = None) -> 'SortQuery':
        return SortQuery(self, col, order, nulls)

    def order_by(self, col: OrderType, order: Order = None, nulls: Nulls = None) -> 'OrderQuery':
        return OrderQuery(self, col, order, nulls)

    def top(self, num_rows: int, col: Column, order: Order = None, nulls: Nulls = None) -> 'TopQuery':
        return TopQuery(self, num_rows, col, order, nulls)

    def join(self, query: 'Query', kind: JoinKind = None):
        return JoinQuery(self, query, kind)

    def project(self, *args: Union[Column, AssignmentBase, BaseExpression], **kwargs: ExpressionType) -> 'ProjectQuery':
        columns: List[Column] = []
        assignments: List[AssignmentBase] = []
        for arg in args:
            if isinstance(arg, Column):
                columns.append(arg)
            elif isinstance(arg, AssignmentBase):
                assignments.append(arg)
            elif isinstance(arg, BaseExpression):
                assignments.append(arg.assign_to())
            else:
                raise ValueError("Invalid assignment: " + arg.to_kql())
        for column_name, expression in kwargs.items():
            assignments.append(AssignmentToSingleColumn(Column(column_name), expression))
        return ProjectQuery(self, columns, assignments)

    def project_rename(self, *args: AssignmentFromColumnToColumn, **kwargs: Column) -> 'ProjectRenameQuery':
        assignments: List[AssignmentFromColumnToColumn] = list(args)
        for column_name, column in kwargs.items():
            assignments.append(AssignmentFromColumnToColumn(Column(column_name), column))
        return ProjectRenameQuery(self, assignments)

    def project_away(self, *columns: StringType):
        return ProjectAwayQuery(self, columns)

    def distinct(self, *columns: BaseExpression):
        return DistinctQuery(self, columns)

    def distinct_all(self):
        return DistinctQuery(self, (AnyExpression(KQL("*")),))

    def extend(self, *args: Union[BaseExpression, AssignmentBase], **kwargs: ExpressionType) -> 'ExtendQuery':
        assignments: List[AssignmentBase] = []
        for arg in args:
            if isinstance(arg, BaseExpression):
                assignments.append(arg.assign_to())
            else:
                assignments.append(arg)
        for column_name, expression in kwargs.items():
            if isinstance(expression, BaseExpression):
                assignments.append(expression.assign_to(Column(column_name)))
            else:
                assignments.append(AnyExpression(to_kql(expression)).assign_to(Column(column_name)))
        return ExtendQuery(self, *assignments)

    def summarize(self, *args: Union[AggregationExpression, AssignmentFromAggregationToColumn],
                  **kwargs: AggregationExpression) -> 'SummarizeQuery':
        assignments: List[AssignmentFromAggregationToColumn] = []
        for arg in args:
            if isinstance(arg, AggregationExpression):
                assignments.append(arg.assign_to())
            elif isinstance(arg, AssignmentFromAggregationToColumn):
                assignments.append(arg)
            else:
                raise ValueError("Invalid assignment: " + str(arg))
        for column_name, agg in kwargs.items():
            assignments.append(AssignmentFromAggregationToColumn(Column(column_name), agg))
        return SummarizeQuery(self, assignments)

    def mv_expand(self, *columns: Column, bag_expansion: BagExpansion = None, with_item_index: Column = None,
                  limit: int = None):
        if len(columns) == 0:
            raise ValueError("Please specify one or more columns for mv-expand")
        return MvExpandQuery(self, columns, bag_expansion, with_item_index, limit)

    def custom(self, custom_query: str):
        return CustomQuery(self, custom_query)

    # TODO convert python types to kusto types
    def evaluate(self, udf: FunctionType, type_spec_str: str):
        return EvaluatePythonQuery(self, udf, type_spec_str)

    @abstractmethod
    def _compile(self) -> KQL:
        pass

    def _compile_all(self, use_full_table_name) -> KQL:
        if self._head is None:
            if self._table is None:
                return KQL("")
            else:
                table = self._table
                if use_full_table_name:
                    return table.get_full_table()
                else:
                    return table.get_table()
        else:
            return KQL("{} | {}".format(self._head._compile_all(use_full_table_name), self._compile()))

    def get_table(self) -> Table:
        if self._head is None:
            return self._table
        else:
            return self._head.get_table()

    def render(self, use_full_table_name: bool = False) -> KQL:
        result = self._compile_all(use_full_table_name)
        logger.debug("Complied query: " + result)
        return result

    def execute(self, table: Table = None):
        if self.get_table() is None:
            if table is None:
                raise RuntimeError("No table supplied")
            rendered_query = table.get_table() + self.render()
        else:
            if table is not None:
                raise RuntimeError("This table is already bound to a query")
            table = self.get_table()
            rendered_query = self.render()

        logger.debug("Running query: " + rendered_query)
        return table.execute(rendered_query)

    def to_dataframe(self, table: Table = None):
        res = self.execute(table)
        return dataframe_from_result_table(res.primary_results[0])


class ProjectQuery(Query):
    _columns: List[Column]
    _assignments: List[AssignmentBase]

    def __init__(self, head: 'Query', columns: List[Column], assignments: List[AssignmentBase]) -> None:
        super().__init__(head)
        self._columns = columns
        self._assignments = assignments

    def _compile(self) -> KQL:
        return KQL('project {}'.format(', '.join(chain(
            (c.kql for c in self._columns),
            (a.to_kql() for a in self._assignments)
        ))))


class ProjectRenameQuery(Query):
    _assignments: List[AssignmentBase]

    def __init__(self, head: 'Query', assignments: List[AssignmentFromColumnToColumn]) -> None:
        super().__init__(head)
        self._assignments = assignments

    def _compile(self) -> KQL:
        return KQL('project-rename {}'.format(', '.join(a.to_kql() for a in self._assignments)))


class ProjectAwayQuery(Query):
    _columns: Tuple[StringType, ...]

    def __init__(self, head: 'Query', columns: Tuple[StringType]) -> None:
        super().__init__(head)
        self._columns = columns

    def _compile(self) -> KQL:
        def col_or_wildcard_to_string(col_or_wildcard):
            return col_or_wildcard.kql if isinstance(col_or_wildcard, Column) else col_or_wildcard

        return KQL('project-away {}'.format(', '.join((col_or_wildcard_to_string(c) for c in self._columns))))


class DistinctQuery(Query):
    _columns: Tuple[BaseExpression, ...]

    def __init__(self, head: 'Query', columns: Tuple[BaseExpression]) -> None:
        super().__init__(head)
        self._columns = columns

    def _compile(self) -> KQL:
        return KQL('distinct {}'.format(', '.join((c.kql for c in self._columns))))


class ExtendQuery(Query):
    _assignments: Tuple[AssignmentBase, ...]

    def __init__(self, head: 'Query', *assignments: AssignmentBase) -> None:
        super().__init__(head)
        self._assignments = assignments

    def _compile(self) -> KQL:
        return KQL('extend {}'.format(', '.join(a.to_kql() for a in self._assignments)))


class WhereQuery(Query):
    _predicate: BooleanType

    def __init__(self, head: Query, predicate: BooleanType):
        super(WhereQuery, self).__init__(head)
        self._predicate = predicate

    def _compile(self) -> KQL:
        return KQL('where {}'.format(self._predicate.kql))


class _SingleNumberQuery(Query):
    _num_rows: int
    _query_name: str

    def __init__(self, head: Query, query_name: str, num_rows: int):
        super(_SingleNumberQuery, self).__init__(head)
        self._query_name = query_name
        self._num_rows = num_rows

    def _compile(self) -> KQL:
        return KQL('{} {}'.format(self._query_name, self._num_rows))


class TakeQuery(_SingleNumberQuery):
    _num_rows: int

    def __init__(self, head: Query, num_rows: int):
        super(TakeQuery, self).__init__(head, 'take', num_rows)


class LimitQuery(_SingleNumberQuery):
    _num_rows: int

    def __init__(self, head: Query, num_rows: int):
        super(LimitQuery, self).__init__(head, 'limit', num_rows)


class SampleQuery(_SingleNumberQuery):
    _num_rows: int

    def __init__(self, head: Query, num_rows: int):
        super(SampleQuery, self).__init__(head, 'sample', num_rows)


class CountQuery(Query):
    _num_rows: int

    def __init__(self, head: Query):
        super(CountQuery, self).__init__(head)

    def _compile(self) -> KQL:
        return KQL('count')


class _OrderQueryBase(Query):
    class OrderSpec:
        col: OrderType
        order: Order
        nulls: Nulls

        def __init__(self, col: OrderType, order: Order, nulls: Nulls):
            self.col = col
            self.order = order
            self.nulls = nulls

    _query_name: str
    _order_specs: List[OrderSpec]

    def __init__(self, head: Query, query_name: str, col: OrderType, order: Order, nulls: Nulls):
        super(_OrderQueryBase, self).__init__(head)
        self._query_name = query_name
        self._order_specs = []
        self.then_by(col, order, nulls)

    def then_by(self, col: OrderType, order: Order = None, nulls: Nulls = None):
        self._order_specs.append(_OrderQueryBase.OrderSpec(col, order, nulls))
        return self

    @staticmethod
    def _compile_order_spec(order_spec):
        res = str(order_spec.col.kql)
        if order_spec.order is not None:
            res += " " + str(order_spec.order.value)
        if order_spec.nulls is not None:
            res += " nulls " + str(order_spec.nulls.value)
        return res

    def _compile(self) -> KQL:
        return KQL(
            '{} by {}'.format(self._query_name,
                              ", ".join([self._compile_order_spec(order_spec) for order_spec in self._order_specs])))


class SortQuery(_OrderQueryBase):
    def __init__(self, head: Query, col: OrderType, order: Order, nulls: Nulls):
        super(SortQuery, self).__init__(head, "sort", col, order, nulls)


class OrderQuery(_OrderQueryBase):
    def __init__(self, head: Query, col: OrderType, order: Order, nulls: Nulls):
        super(OrderQuery, self).__init__(head, "order", col, order, nulls)


class TopQuery(Query):
    _num_rows: int
    _order_spec: OrderQuery.OrderSpec

    def __init__(self, head: Query, num_rows: int, col: Column, order: Order, nulls: Nulls):
        super(TopQuery, self).__init__(head)
        self._num_rows = num_rows
        self._order_spec = OrderQuery.OrderSpec(col, order, nulls)

    def _compile(self) -> KQL:
        # noinspection PyProtectedMember
        return KQL('top {} by {}'.format(self._num_rows, SortQuery._compile_order_spec(self._order_spec)))


class JoinException(Exception):
    pass


class JoinQuery(Query):
    _joined_query: Query
    _kind: JoinKind
    _on_attributes: Tuple[Tuple[Column, ...], ...]

    def __init__(self, head: Query, joined_query: Query, kind: JoinKind,
                 on_attributes: Tuple[Tuple[Column, ...], ...] = tuple()):
        super(JoinQuery, self).__init__(head)
        self._joined_query = joined_query
        self._kind = kind
        self._on_attributes = on_attributes

    def on(self, col1: Column, col2: Column = None) -> 'JoinQuery':
        self._on_attributes = self._on_attributes + (((col1,),) if col2 is None else ((col1, col2),))
        return self

    @staticmethod
    def _compile_on_attribute(attribute: Tuple[Column]):
        assert len(attribute) in (1, 2)
        if len(attribute) == 1:
            return attribute[0].kql
        else:
            return "$left.{}==$right.{}".format(attribute[0].kql, attribute[1].kql)

    def _compile(self) -> KQL:
        if len(self._on_attributes) == 0:
            raise JoinException("A call to join() must be followed by a call to on()")
        if self._joined_query.get_table() is None:
            raise JoinException("The joined query must have a table")

        return KQL("join {} ({}) on {}".format(
            "" if self._kind is None else "kind={}".format(self._kind.value),
            self._joined_query.render(use_full_table_name=True),
            ", ".join([self._compile_on_attribute(attr) for attr in self._on_attributes])))


class SummarizeQuery(Query):
    _assignments: List[AssignmentFromAggregationToColumn]
    _by_columns: List[Union[Column, BaseExpression]]
    _by_assignments: List[AssignmentToSingleColumn]

    def __init__(self, head: Query,
                 assignments: List[AssignmentFromAggregationToColumn]):
        super(SummarizeQuery, self).__init__(head)
        self._assignments = assignments
        self._by_columns = []
        self._by_assignments = []

    def by(self, *args: Union[AssignmentToSingleColumn, Column, BaseExpression],
           **kwargs: BaseExpression):
        for arg in args:
            if isinstance(arg, Column) or isinstance(arg, BaseExpression):
                self._by_columns.append(arg)
            elif isinstance(arg, AssignmentToSingleColumn):
                self._by_assignments.append(arg)
            else:
                raise ValueError("Invalid assignment: " + arg.to_kql())
        for column_name, group_exp in kwargs.items():
            self._by_assignments.append(AssignmentToSingleColumn(Column(column_name), group_exp))
        return self

    def _compile(self) -> KQL:
        result = 'summarize {}'.format(', '.join(a.to_kql() for a in self._assignments))
        if len(self._by_assignments) != 0 or len(self._by_columns) != 0:
            result += ' by {}'.format(', '.join(chain(
                (c.kql for c in self._by_columns),
                (a.to_kql() for a in self._by_assignments)
            )))
        return KQL(result)


class MvExpandQuery(Query):
    _columns: Tuple[Column]
    _bag_expansion: BagExpansion
    _with_item_index: Column
    _limit: int

    def __init__(self, head: Query, columns: Tuple[Column], bag_expansion: BagExpansion, with_item_index: Column,
                 limit: int):
        super(MvExpandQuery, self).__init__(head)
        self._columns = columns
        self._bag_expansion = bag_expansion
        self._with_item_index = with_item_index
        self._limit = limit

    def _compile(self) -> KQL:
        res = "mv-expand "
        if self._bag_expansion is not None:
            res += "bagexpansion={} ".format(self._bag_expansion.value)
        if self._with_item_index is not None:
            res += "with_itemindex={} ".format(self._with_item_index.kql)
        res += ", ".join([c.kql for c in self._columns])
        if self._limit:
            res += " limit {}".format(self._limit)
        return KQL(res)


class CustomQuery(Query):
    _custom_query: str

    def __init__(self, head: Query, custom_query: str):
        super(CustomQuery, self).__init__(head)
        self._custom_query = custom_query

    def _compile(self) -> KQL:
        return KQL(self._custom_query)


class EvaluatePythonQuery(Query):
    _udf: FunctionType
    _type_specs: str

    def __init__(self, head: Query, udf: FunctionType, type_specs: str):
        super(EvaluatePythonQuery, self).__init__(head)
        self._udf = udf
        self._type_specs = type_specs

    def _compile(self) -> KQL:
        return KQL('evaluate python({},"{}")'.format(
            self._type_specs,
            stringify_python_func(self._udf)
        ))



from datetime import timedelta

from pykusto.expressions import column_generator as col
from pykusto.query import Query
from test.test_base import TestBase


class TestExpressions(TestBase):
    def test_contains(self):
        self.assertEqual(
            ' | where foo contains "bar"',
            Query().where(col.foo.contains('bar')).render(),
        )
        self.assertEqual(
            ' | where foo contains_cs "bar"',
            Query().where(col.foo.contains('bar', True)).render(),
        )

    def test_not_contains(self):
        self.assertEqual(
            ' | where foo !contains "bar"',
            Query().where(col.foo.not_contains('bar')).render(),
        )
        self.assertEqual(
            ' | where foo !contains_cs "bar"',
            Query().where(col.foo.not_contains('bar', True)).render(),
        )

    def test_array_access(self):
        self.assertEqual(
            ' | where (arr[3]) == "bar"',
            Query().where(col.arr[3] == 'bar').render(),
        )

    def test_array_access_expression_index(self):
        self.assertEqual(
            ' | where (arr[(foo * 2)]) == "bar"',
            Query().where(col.arr[col.foo * 2] == 'bar').render(),
        )

    def test_array_access_yields_any_expression(self):
        self.assertEqual(
            ' | where (cos(arr[3])) < 1',
            Query().where(col.arr[3].cos() < 1).render(),
        )

    def test_mapping_access(self):
        self.assertEqual(
            ' | where (dict["key"]) == "bar"',
            Query().where(col.dict['key'] == 'bar').render(),
        )

    def test_mapping_access_attribute(self):
        self.assertEqual(
            ' | where (dict.key) == "bar"',
            Query().where(col.dict.key == 'bar').render(),
        )

    def test_mapping_access_expression_index(self):
        self.assertEqual(
            ' | where (dict[foo]) == "bar"',
            Query().where(col.dict[col.foo] == 'bar').render(),
        )

    def test_mapping_access_yields_any_expression(self):
        self.assertEqual(
            ' | where (dict["key"]) contains "substr"',
            Query().where(col.dict['key'].contains("substr")).render(),
        )

    def test_dynamic(self):
        self.assertEqual(
            ' | where (dict["foo"][0].bar[1][2][(tolower(bar))]) > time(1.0:0:0.0)',
            Query().where(col.dict['foo'][0].bar[1][2][col.bar.lower()] > timedelta(1)).render(),
        )

    def test_assign_to(self):
        self.assertEqual(
            " | extend bar = (foo * 2)",
            Query().extend((col.foo * 2).assign_to(col.bar)).render(),
        )
        self.assertEqual(
            " | extend foo = (shoo * 2)",
            Query().extend(foo=(col.shoo * 2)).render(),
        )

    def test_extend_const(self):
        self.assertEqual(
            " | extend foo = (5), bar = (\"bar\"), other_col = other",
            Query().extend(foo=5, bar="bar", other_col=col.other).render(),
        )

    def test_between_timespan(self):
        self.assertEqual(
            " | where foo between (time(0.0:0:0.0) .. time(0.3:0:0.0))",
            Query().where(col.foo.between(timedelta(0), timedelta(hours=3))).render(),
        )

    def test_is_empty(self):
        self.assertEqual(
            'isempty(foo)',
            col.foo.is_empty().kql,
        )

    def test_column_generator(self):
        self.assertEqual(
            " | project ['foo.bar']",
            Query().project(col['foo.bar']).render(),
        )

    def test_is_in(self):
        self.assertEqual(
            " | where foo in (\"A\", \"B\", \"C\")",
            Query().where(col.foo.is_in(["A", "B", "C"])).render()
        )
        self.assertEqual(
            " | where foo in (\"[\", \"[[\", \"]\")",
            Query().where(col.foo.is_in(['[', "[[", "]"])).render()
        )

    def test_has(self):
        self.assertEqual(
            " | where foo has \"test\"",
            Query().where(col.foo.has("test")).render()
        )


import json
from typing import Union

from pykusto.expressions import Column, NumberType, NumberExpression, TimespanType, \
    DatetimeExpression, TimespanExpression, ArrayType, DynamicType, DatetimeType, BaseExpression, BooleanType, \
    ExpressionType, AggregationExpression, StringType, StringExpression, BooleanExpression, \
    NumberAggregationExpression, MappingAggregationExpression, ArrayAggregationExpression, to_kql, DynamicExpression, \
    ArrayExpression
from pykusto.kql_converters import KQL
from pykusto.type_utils import plain_expression


# Scalar functions
def acos(expr: NumberType) -> NumberExpression:
    return expr.acos()


def ago(expr: TimespanType) -> DatetimeExpression:
    return TimespanExpression.ago(expr)


# def array_concat(): return
#
#
# def array_iif(): return


def array_length(expr: ArrayType) -> NumberExpression:
    return expr.array_length()


# def array_slice(): return
#
#
# def array_split(): return
#
#
# def asin(): return
#
#
# def atan(self): return
#
#
# def atan2(self): return
#
#
# def base64_decode_toarray(self): return
#
#
# def base64_decode_tostring(self): return
#
#
# def base64_encode_tostring(self): return
#
#
def bag_keys(expr: DynamicType):
    return expr.keys()


# def beta_cdf(self): return
#
#
# def beta_inv(self): return
#
#
# def beta_pdf(self): return


def bin(expr: Union[NumberType, DatetimeType, TimespanType],
        round_to: Union[NumberType, TimespanType]) -> BaseExpression:
    """
    Refers only to bin() as part of summarize by bin(...),
     if you wish to use it as a scalar function, use 'floor()' instead
    :param expr: A number, date, or timespan.
    :param round_to: The "bin size". A number, date or timespan that divides value.
    :return:
    """
    return expr.bin(round_to)


def bin_at(expr: Union[NumberType, DatetimeType, TimespanType],
           bin_size: Union[NumberType, TimespanType],
           fixed_point: Union[NumberType, DatetimeType, TimespanType]) -> BaseExpression:
    return expr.bin_at(bin_size, fixed_point)


def bin_auto(expr: Union[NumberType, DatetimeType, TimespanType]) -> BaseExpression:
    return expr.bin_auto()


# def binary_and(self): return
#
#
# def binary_not(self): return
#
#
# def binary_or(self): return
#
#
# def binary_shift_left(self): return
#
#
# def binary_shift_right(self): return
#
#
# def binary_xor(self): return


def case(predicate: BooleanType, val: ExpressionType, *args: Union[BooleanType, ExpressionType]) -> BaseExpression:
    res = 'case({}, {}, {})'.format(
        to_kql(predicate), to_kql(val), ', '.join([to_kql(arg) for arg in args])
    )
    return AggregationExpression(KQL(res))


def ceiling(expr: NumberType) -> NumberExpression:
    return expr.ceiling()


# def coalesce(self): return
#
#
# def column_ifexists(self): return
#
#

def cos(expr: NumberType) -> NumberExpression:
    return expr.cos()


#
# def cot(self): return

# def countof(self): return
#
#
# def current_cluster_endpoint(self): return
#
#
# def current_cursor(self): return
#
#
# def current_database(self): return
#
#
# def current_principal(self): return
#
#
# def cursor_after(self): return
#
#
# def cursor_before_or_at(self): return
#
#
# def cursor_current(self): return
#
# def datetime_add(self): return
#
#
# def datetime_part(self): return
#
#
# def datetime_diff(self): return
#
#
# def dayofmonth(self): return
#
#
# def dayofweek(self): return
#
#
# def dayofyear(self): return


# def dcount_hll(expr: ExpressionType) -> BaseExpression:
#     return BaseExpression(KQL('dcount_hll({})'.format(to_kql(expr))))


# def degrees(self): return


def endofday(expr: DatetimeExpression, offset: NumberType = None) -> DatetimeExpression:
    return expr.endofday(offset)


def endofmonth(expr: DatetimeType, offset: NumberType = None) -> DatetimeExpression:
    return expr.endofmonth(offset)


def endofweek(expr: DatetimeType, offset: NumberType = None) -> DatetimeExpression:
    return expr.endofweek(offset)


def endofyear(expr: DatetimeType, offset: NumberType = None) -> DatetimeExpression:
    return expr.endofyear(offset)


# def estimate_data_size(self): return


def exp(expr: NumberType) -> NumberExpression:
    return expr.exp()


def exp10(expr: NumberType) -> NumberExpression:
    return expr.exp10()


def exp2(expr: NumberType) -> NumberExpression:
    return expr.exp2()


# def extent_id(self): return
#
#
# def extent_tags(self): return
#
#
# def extract(self): return
#
#
# def extract_all(self): return
#
#
# def extractjson(self): return


def floor(expr: Union[NumberType, DatetimeType],
          round_to: Union[NumberType, TimespanType]) -> Union[NumberExpression, DatetimeExpression]:
    return expr.floor(round_to)


def format_datetime(expr: DatetimeExpression, format_string: StringType) -> StringExpression:
    return expr.format_datetime(format_string)


def format_timespan(expr: TimespanType, format_string: StringType) -> StringExpression:
    return expr.format_timespan(format_string)


# def gamma(self): return


def getmonth(expr: DatetimeType) -> NumberExpression:
    return expr.getmonth()


def gettype(expr: ExpressionType) -> StringExpression:
    return expr.gettype()


def getyear(expr: DatetimeType) -> NumberExpression:
    return expr.getyear()


def hash(expr: ExpressionType):
    return expr.__hash__()


def hash_sha256(expr: ExpressionType):
    return expr.hash_sha256()


def hourofday(expr: DatetimeType) -> NumberExpression:
    return expr.hourofday()


def iff(predicate: BooleanType, if_true: ExpressionType, if_false: ExpressionType) -> BaseExpression:
    return_type = type(if_true)
    if type(if_false) is not return_type:
        raise TypeError("The second and third arguments must be of the same type")
    return plain_expression.for_type(return_type)(
        KQL('iff({}, {}, {})'.format(to_kql(predicate), to_kql(if_true), to_kql(if_false)))
    )


def iif(predicate: BooleanType, if_true: ExpressionType, if_false: ExpressionType) -> BaseExpression:
    return iff(predicate, if_true, if_false)


#
# def indexof(self): return
#
#
# def indexof_regex(self): return
#
#
# def ingestion_time(self): return
#
#
# def isascii(self): return


def isempty(expr: ExpressionType) -> BooleanExpression:
    return expr.is_empty()


def isfinite(expr: NumberType) -> BooleanExpression:
    return expr.isfinite()


def isinf(expr: NumberType) -> BooleanExpression:
    return expr.isinf()


def isnan(expr: NumberExpression) -> BooleanExpression:
    return expr.isnan()


def isnotempty(expr: ExpressionType) -> BooleanExpression:
    return expr.is_not_empty()


def isnotnull(expr: ExpressionType) -> BooleanExpression:
    return expr.is_not_null()


def isnull(expr: ExpressionType) -> BooleanExpression:
    return expr.is_null()


def isutf8(expr: StringType) -> BooleanExpression:
    return expr.is_utf8()


def log(expr: NumberType) -> NumberExpression:
    return expr.log()


def log10(expr: NumberType) -> NumberExpression:
    return expr.log10()


def log2(expr: NumberType) -> NumberExpression:
    return expr.log2()


def loggamma(expr: NumberType) -> NumberExpression:
    return expr.loggamma()


def make_datetime(year: NumberType,
                  month: NumberType,
                  day: NumberType,
                  hour: NumberType = None,
                  minute: NumberType = None,
                  second: NumberType = None) -> DatetimeExpression:
    res = 'make_datetime({year}, {month}, {day}, {hour}, {minute}, {second})'.format(
        year=to_kql(year),
        month=to_kql(month),
        day=to_kql(day),
        hour=to_kql(0 if hour is None else hour),
        minute=to_kql(0 if minute is None else minute),
        second=to_kql(0 if second is None else second)
    )
    return DatetimeExpression(KQL(res))


def make_string(): raise NotImplemented  # TODO


def make_timespan(): raise NotImplemented  # TODO


# def max_of(self): return
#
#
# def min_of(self): return


def monthofyear(): raise NotImplemented  # TODO


def new_guid(): raise NotImplemented  # TODO


def now(offset: TimespanType = None) -> StringExpression:
    if offset:
        return StringExpression(KQL('now({})'.format(to_kql(offset))))
    return StringExpression(KQL('now()'))


def pack(): raise NotImplemented  # TODO


def pack_all(): raise NotImplemented  # TODO


def pack_array(): raise NotImplemented  # TODO


def pack_dictionary(): raise NotImplemented  # TODO


#
#
# def parse_csv(self): return
#
#
# def parse_ipv4(self): return


def parse_json(expr: Union[StringType, DynamicType]) -> DynamicExpression:
    return DynamicExpression(KQL('parse_json({})'.format(to_kql(expr))))


# def parse_path(self): return
#
#
# def parse_url(self): return
#
#
# def parse_urlquery(self): return
#
#
# def parse_user_agent(self): return
#
#
# def parse_version(self): return
#
#
# def parse_xml(self): return


def percentile_tdigest(): raise NotImplemented  # TODO


def percentrank_tdigest(): raise NotImplemented  # TODO


def pow(expr1: NumberType, expr2: NumberType) -> NumberExpression:
    return NumberExpression(KQL('pow({}, {})'.format(to_kql(expr1), to_kql(expr2))))


# def radians(self): return
#
#
# def rand(self): return
#
#
# def range(self): return
#
#
# def rank_tdigest(self): return
#
#
# def repeat(self): return


# def replace(self): return


# def reverse(self): return


def round(expr: NumberType, precision: NumberType = None) -> NumberExpression:
    return expr.round(precision)


# def series_add(self): return
#
#
# def series_decompose(self): return
#
#
# def series_decompose_anomalies(self): return
#
#
# def series_decompose_forecast(self): return
#
#
# def series_divide(self): return
#
#
# def series_equals(self): return
#
#
# def series_fill_backward(self): return
#
#
# def series_fill_const(self): return
#
#
# def series_fill_forward(self): return
#
#
# def series_fill_linear(self): return
#
#
# def series_fir(self): return
#
#
# def series_fit_2lines(self): return
#
#
# def series_fit_2lines_dynamic(self): return
#
#
# def series_fit_line(self): return
#
#
# def series_fit_line_dynamic(self): return
#
#
# def series_greater(self): return
#
#
# def series_greater_equals(self): return
#
#
# def series_iir(self): return
#
#
# def series_less(self): return
#
#
# def series_less_equals(self): return
#
#
# def series_multiply(self): return
#
#
# def series_not_equals(self): return
#
#
# def series_outliers(self): return
#
#
# def series_periods_detect(self): return
#
#
# def series_periods_validate(self): return
#
#
# def series_seasonal(self): return
#
#
# def series_stats(self): return
#
#
# def series_stats_dynamic(self): return
#
#
# def series_subtract(self): return
#
#
# def set_difference(self): return
#
#
# def set_intersect(self): return
#
#
# def set_union(self): return


def sign(expr: NumberType) -> NumberExpression:
    return NumberExpression(KQL('sign({})'.format(to_kql(expr))))


# def sin(self): return
#
#
def split(string: StringType, delimiter: StringType, requested_index: NumberType = None) -> 'ArrayExpression':
    return StringExpression(KQL(string)).split(delimiter, requested_index)


def sqrt(expr: NumberType) -> NumberExpression:
    return NumberExpression(KQL('sqrt({})'.format(to_kql(expr))))


def startofday(expr: DatetimeType, offset: NumberType = None) -> DatetimeExpression:
    return expr.startofday(offset)


def startofmonth(expr: DatetimeType, offset: NumberType = None) -> DatetimeExpression:
    return expr.startofmonth(offset)


def startofweek(expr: DatetimeType, offset: NumberType = None) -> DatetimeExpression:
    return expr.startofweek(offset)


def startofyear(expr: DatetimeType, offset: NumberType = None) -> DatetimeExpression:
    return expr.startofyear(offset)


def strcat(*strings: StringType) -> StringExpression:
    if len(strings) < 2:
        raise ValueError("strcat requires at least two arguments")
    return StringExpression(KQL('strcat({})'.format(', '.join(to_kql(s) for s in strings))))


def to_literal_dynamic(d: DynamicType) -> KQL:
    if isinstance(d, BaseExpression):
        return d.kql
    return KQL('dynamic({})'.format(json.dumps(d)))


def strcat_array(expr: ArrayType, delimiter: StringType) -> StringExpression:
    return StringExpression(KQL('strcat_array({}, {})'.format(to_literal_dynamic(expr), to_kql(delimiter))))


def strcat_delim(delimiter: StringType, expr1: StringType, expr2: StringType, *exprs: StringType) -> StringExpression:
    res = 'strcat_delim({}, {}, {}'.format(
        to_kql(delimiter),
        to_kql(expr1),
        to_kql(expr2))
    if len(exprs) > 0:
        res = res + ', ' + ', '.join([to_kql(expr) for expr in exprs])
    return StringExpression(KQL(res + ')'))


def strcmp(expr1: StringType, expr2: StringType) -> NumberExpression:
    return NumberExpression(KQL('strcmp({}, {})'.format(to_kql(expr1), to_kql(expr2))))


def string_size(expr: StringType) -> NumberExpression:
    return NumberExpression(KQL('string_size({})'.format(to_kql(expr))))


def strlen(expr: StringType) -> NumberExpression:
    return NumberExpression(KQL('strlen({})'.format(to_kql(expr))))


def strrep(expr: StringType,
           multiplier: NumberType,
           delimiter: StringType = None) -> StringExpression:
    if delimiter is None:
        res = 'strrep({}, {})'.format(to_kql(expr), to_kql(multiplier))
    else:
        res = 'strrep({}, {}, {})'.format(to_kql(expr), to_kql(multiplier),
                                          to_kql(delimiter))
    return StringExpression(KQL((res)))


def substring(expr: StringType, start_index: NumberType, length: NumberType = None) -> StringExpression:
    return StringExpression(KQL(
        ('substring({}, {})' if length is None else 'substring({}, {}, {})').format(
            to_kql(expr), to_kql(start_index), to_kql(length)
        )
    ))


# def tan(self): return
#
#
# def tdigest_merge(self): return


def tobool(expr: ExpressionType) -> BooleanExpression:
    return BooleanExpression(KQL('tobool({})'.format(to_kql(expr))))


def toboolean(expr: ExpressionType) -> BooleanExpression:
    return BooleanExpression(KQL('toboolean({})'.format(to_kql(expr))))


def todatetime(expr: StringType) -> DatetimeExpression:
    return DatetimeExpression(KQL('todatetime({})'.format(to_kql(expr))))


def todecimal(): raise NotImplemented  # TODO


def todouble(expr: NumberType) -> NumberExpression:
    return NumberExpression(KQL("todouble({})".format(to_kql(expr))))


def todynamic(): raise NotImplemented  # TODO


def toguid(): raise NotImplemented  # TODO


def tohex(): raise NotImplemented  # TODO


def toint(): raise NotImplemented  # TODO


def tolong(): raise NotImplemented  # TODO


def tolower(): raise NotImplemented  # TODO


def toreal(): raise NotImplemented  # TODO


def tostring(expr: ExpressionType):
    return expr.to_string()


def totimespan(): raise NotImplemented  # TODO


def toupper(): raise NotImplemented  # TODO


# def to_utf8(self): return
#
#
# def translate(self): return
#
#
# def treepath(self): return


def trim(): raise NotImplemented  # TODO


# def trim_end(self): return
#
#
# def trim_start(self): return


def url_decode(): raise NotImplemented  # TODO


def url_encode(): raise NotImplemented  # TODO


def weekofyear(): raise NotImplemented  # TODO


# def welch_test(self): return


def zip(): raise NotImplemented  # TODO


# ----------------------------------------------------
# aggregative functions
# -----------------------------------------------------

def any(*args: ExpressionType) -> AggregationExpression:
    res = 'any({})'.format(', '.join([arg.kql for arg in args]))
    return AggregationExpression(KQL(res))


def arg_max(*args: ExpressionType) -> AggregationExpression:
    res = 'arg_max({})'.format(', '.join([arg.kql for arg in args]))
    return AggregationExpression(KQL(res))


def arg_min(*args: ExpressionType) -> AggregationExpression:
    res = 'arg_min({})'.format(', '.join([arg.kql for arg in args]))
    return AggregationExpression(KQL(res))


def avg(expr: ExpressionType) -> NumberAggregationExpression:
    return NumberAggregationExpression(KQL('avg({})'.format(to_kql(expr))))


def avgif(expr: ExpressionType, predicate: BooleanType) -> NumberAggregationExpression:
    return NumberAggregationExpression(KQL('avgif({}, {})'.format(to_kql(expr), to_kql(predicate))))


# def buildschema(self):
#     return


def count(col: Column = None) -> NumberAggregationExpression:
    res = "count()" if col is None else "count({})".format(col.kql)
    return NumberAggregationExpression(KQL(res))


def countif(predicate: BooleanType) -> NumberAggregationExpression:
    return NumberAggregationExpression(KQL('countif({})'.format(to_kql(predicate))))


def dcount(expr: ExpressionType, accuracy: NumberType = None) -> NumberAggregationExpression:
    return NumberAggregationExpression(KQL(
        ('dcount({})' if accuracy is None else 'dcount({}, {})').format(to_kql(expr), to_kql(accuracy))
    ))


def dcountif(expr: ExpressionType, predicate: BooleanType, accuracy: NumberType = 0) -> NumberAggregationExpression:
    return NumberAggregationExpression(KQL('dcountif({}, {}, {})'.format(
        to_kql(expr), to_kql(predicate), to_kql(accuracy)
    )))


# def hll(expr: ExpressionType, accuracy: NumberType = None) -> AggregationExpression:
#     return AggregationExpression(KQL(
#         ('hll({})' if accuracy is None else 'hll({}, {})').format(expr, accuracy)
#     ))


# def hll_merge(expr: ExpressionType) -> AggregationExpression:
#     return AggregationExpression(KQL('hll_merge({})'.format(to_kql(expr))))


def make_bag(expr: ExpressionType, max_size: NumberType = None) -> MappingAggregationExpression:
    if max_size:
        return MappingAggregationExpression(KQL('make_bag({}, {})'.format(expr, max_size)))
    return MappingAggregationExpression(KQL('make_bag({})'.format(to_kql(expr))))


def make_list(expr: ExpressionType, max_size: NumberType = None) -> ArrayAggregationExpression:
    if max_size:
        return ArrayAggregationExpression(KQL('make_list({}, {})'.format(expr, max_size)))
    return ArrayAggregationExpression(KQL('make_list({})'.format(to_kql(expr))))


def make_set(expr: ExpressionType, max_size: NumberType = None) -> ArrayAggregationExpression:
    if max_size:
        return ArrayAggregationExpression(KQL('make_set({}, {})'.format(expr, max_size)))
    return ArrayAggregationExpression(KQL('make_set({})'.format(to_kql(expr))))


def max(expr: ExpressionType) -> AggregationExpression:
    return AggregationExpression(KQL('max({})'.format(to_kql(expr))))


def min(expr: ExpressionType) -> AggregationExpression:
    return AggregationExpression(KQL('min({})'.format(to_kql(expr))))


def percentile(expr: ExpressionType, per: NumberType) -> AggregationExpression:
    res = 'percentiles({}, {})'.format(expr, to_kql(per))
    return AggregationExpression(KQL(res))


def percentiles(expr: ExpressionType, *pers: NumberType) -> AggregationExpression:
    res = 'percentiles({}, {})'.format(expr,
                                       ', '.join([str(to_kql(per)) for per in pers]))
    return AggregationExpression(KQL(res))


def stdev(expr: ExpressionType) -> AggregationExpression:
    return AggregationExpression(KQL('stdev({})'.format(to_kql(expr))))


def stdevif(expr: ExpressionType, predicate: BooleanType) -> AggregationExpression:
    return AggregationExpression(KQL('stdevif({}, {})'.format(to_kql(expr), to_kql(predicate))))


def stdevp(expr: ExpressionType) -> AggregationExpression:
    return AggregationExpression(KQL('stdevp({})'.format(to_kql(expr))))


def sum(expr: ExpressionType) -> AggregationExpression:
    return AggregationExpression(KQL('sum({})'.format(to_kql(expr))))


def sumif(expr: ExpressionType, predicate: BooleanType) -> AggregationExpression:
    return AggregationExpression(KQL('sumif({}, {})'.format(to_kql(expr), to_kql(predicate))))


# def tdigest(self):
#     return
#
#
# def tdigest_merge(self):
#     return


def variance(expr: ExpressionType) -> AggregationExpression:
    return AggregationExpression(KQL('variance({})'.format(to_kql(expr))))


def varianceif(expr: ExpressionType, predicate: BooleanType) -> AggregationExpression:
    return AggregationExpression(KQL('varianceif({}, {})'.format(to_kql(expr), to_kql(predicate))))


def variancep(expr: ExpressionType) -> AggregationExpression:
    return AggregationExpression(KQL('variancep({})'.format(to_kql(expr))))



import json
from datetime import datetime, timedelta
from numbers import Number
from typing import NewType, Mapping, Union, List, Tuple

from pykusto.type_utils import kql_converter

KQL = NewType('KQL', str)


@kql_converter(datetime)
def datetime_to_kql(dt: datetime) -> KQL:
    return KQL(dt.strftime('datetime(%Y-%m-%d %H:%M:%S.%f)'))


@kql_converter(timedelta)
def timedelta_to_kql(td: timedelta) -> KQL:
    hours, remainder = divmod(td.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    return KQL('time({days}.{hours}:{minutes}:{seconds}.{microseconds})'.format(
        days=td.days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        microseconds=td.microseconds,
    ))


@kql_converter(Mapping, List, Tuple)
def dynamic_to_kql(d: Union[Mapping, List, Tuple]) -> KQL:
    query = list(json.dumps(d))
    # Convert square brackets to round brackets (Issue #11)
    counter = 0
    prev = ""
    for i, c in enumerate(query):
        if counter == 0:
            if c == "[":
                query[i] = "("
            elif c == "]":
                query[i] = ")"
            elif c in ['"', '\''] and prev != "\\":
                counter += 1
        elif counter > 0:
            if c in ['"', '\''] and prev != "\\":
                counter -= 1
        prev = query[i]
    assert counter == 0
    return KQL("".join(query))


@kql_converter(bool)
def bool_to_kql(b: bool) -> KQL:
    return KQL('true') if b else KQL('false')


@kql_converter(str)
def str_to_kql(s: str) -> KQL:
    return KQL('"{}"'.format(s))


@kql_converter(Number)
def number_to_kql(n: Number) -> KQL:
    return KQL(str(n))


@kql_converter(type(None))
def none_to_kql(n: type(None)) -> KQL:
    return KQL("")



from types import FunctionType


def stringify_python_func(func: FunctionType):
    return """from types import CodeType\\ncode=CodeType({},{},{},{},{},{},{},{},{},{},{},{},{},{},{})\\nexec(code)\\n""".format(
        func.__code__.co_argcount,
        func.__code__.co_kwonlyargcount,
        func.__code__.co_nlocals,
        func.__code__.co_stacksize,
        func.__code__.co_flags,
        func.__code__.co_code,
        func.__code__.co_consts,
        func.__code__.co_names,
        func.__code__.co_varnames,
        "'{}'".format(func.__code__.co_filename),
        "'{}'".format(func.__code__.co_name),
        func.__code__.co_firstlineno,
        func.__code__.co_lnotab,
        func.__code__.co_freevars,
        func.__code__.co_cellvars
    ).replace("\\x", "\\\\x")



from pykusto import functions as f
from pykusto.client import PyKustoClient
from pykusto.expressions import column_generator as col
from pykusto.query import Query, Order, Nulls, JoinKind, JoinException, BagExpansion
from test.test_base import TestBase
from test.test_table import MockKustoClient


class TestQuery(TestBase):
    def test_sanity(self):
        # test concatenation #
        self.assertEqual(
            " | where foo > 4 | take 5 | sort by bar asc nulls last",
            Query().where(col.foo > 4).take(5).sort_by(col.bar, Order.ASC, Nulls.LAST).render(),
        )

    def test_add_queries(self):
        query_a = Query().where(col.foo > 4)
        query_b = Query().take(5)
        query_c = Query().where(col.foo > 1).sort_by(col.bar, Order.ASC, Nulls.LAST)
        query = query_a + query_b + query_c

        self.assertEqual(
            " | where foo > 4 | take 5 | where foo > 1 | sort by bar asc nulls last",
            query.render(),
        )

        # make sure the originals didn't change
        self.assertEqual(
            " | where foo > 4",
            query_a.render(),
        )
        self.assertEqual(
            " | take 5",
            query_b.render(),
        )
        self.assertEqual(
            " | where foo > 1 | sort by bar asc nulls last",
            query_c.render(),
        )

    def test_add_queries_with_table(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table']
        query_a = Query(table).where(col.foo > 4)
        query_b = Query().take(5).take(2).sort_by(col.bar, Order.ASC, Nulls.LAST)
        query = query_a + query_b
        self.assertEqual(
            "test_table | where foo > 4 | take 5 | take 2 | sort by bar asc nulls last",
            query.render(),
        )

        # make sure the originals didn't change
        self.assertEqual(
            "test_table | where foo > 4",
            query_a.render(),
        )
        self.assertEqual(
            " | take 5 | take 2 | sort by bar asc nulls last",
            query_b.render(),
        )

    def test_where(self):
        self.assertEqual(
            " | where foo > 4",
            Query().where(col.foo > 4).render(),
        )

    def test_take(self):
        self.assertEqual(
            " | take 3",
            Query().take(3).render(),
        )

    def test_sort(self):
        self.assertEqual(
            " | sort by foo desc nulls first",
            Query().sort_by(col.foo, order=Order.DESC, nulls=Nulls.FIRST).render(),
        )

    def test_order(self):
        self.assertEqual(
            " | order by foo desc nulls first",
            Query().order_by(col.foo, order=Order.DESC, nulls=Nulls.FIRST).render(),
        )

    def test_order_expression_in_arg(self):
        self.assertEqual(
            " | order by strlen(foo) desc nulls first",
            Query().order_by(f.strlen(col.foo), order=Order.DESC, nulls=Nulls.FIRST).render(),
        )

    def test_sort_multiple_cols(self):
        self.assertEqual(
            " | sort by foo desc nulls first, bar asc nulls last",
            Query().sort_by(col.foo, order=Order.DESC, nulls=Nulls.FIRST).then_by(col.bar, Order.ASC,
                                                                                  Nulls.LAST).render(),
        )

    def test_no_params_for_sort(self):
        self.assertEqual(
            " | sort by foo, bar",
            Query().sort_by(col.foo).then_by(col.bar).render(),
        )
        self.assertEqual(
            " | sort by foo desc nulls first, bar",
            Query().sort_by(col.foo, order=Order.DESC, nulls=Nulls.FIRST).then_by(col.bar).render(),
        )

    def test_top(self):
        self.assertEqual(
            " | top 3 by foo desc nulls first",
            Query().top(3, col.foo, order=Order.DESC, nulls=Nulls.FIRST).render(),
        )

    def test_join_with_table(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table']

        self.assertEqual(
            ' | where foo > 4 | take 5 | join kind=inner (cluster("test_cluster.kusto.windows.net").database("test_db").table("test_table")) on col0, $left.col1==$right.col2',
            Query().where(col.foo > 4).take(5).join(
                Query(table), kind=JoinKind.INNER).on(col.col0).on(col.col1, col.col2).render(),
        )

    def test_join_with_table_and_query(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table']

        self.assertEqual(
            ' | where foo > 4 | take 5 | join kind=inner (cluster("test_cluster.kusto.windows.net").database("test_db").table("test_table") | where bla == 2 | take 6) on col0, '
            "$left.col1==$right.col2",
            Query().where(col.foo > 4).take(5).join(
                Query(table).where(col.bla == 2).take(6), kind=JoinKind.INNER).on(col.col0).on(col.col1,
                                                                                               col.col2).render(),
        )

    def test_join_no_joined_table(self):
        self.assertRaises(
            JoinException("The joined query must have a table"),
            Query().where(col.foo > 4).take(5).join(
                Query().take(2), kind=JoinKind.INNER).on(col.col0).on(col.col1, col.col2).render
        )

    def test_join_no_on(self):
        self.assertRaises(
            JoinException("A call to join() must be followed by a call to on()"),
            Query().where(col.foo > 4).take(5).join(
                Query().take(2), kind=JoinKind.INNER).render
        )

    def test_extend(self):
        self.assertEqual(
            " | extend sum = (v1 + v2), foo = (bar * 4) | take 5",
            Query().extend((col.v1 + col.v2).assign_to(col.sum), foo=col.bar * 4).take(5).render(),
        )

    def test_extend_assign_to_multiple_columns(self):
        self.assertEqual(
            " | extend (foo1, foo2) = foo, shoo = (bar * 4)",
            Query().extend(col.foo.assign_to(col.foo1, col.foo2), shoo=col.bar * 4).render(),
        )

    def test_extend_generate_column_name(self):
        self.assertEqual(
            " | extend (v1 + v2), foo = (bar * 4)",
            Query().extend(col.v1 + col.v2, foo=col.bar * 4).render(),
        )

    def test_summarize(self):
        self.assertEqual(
            " | summarize count(foo), my_count = count(bar)",
            Query().summarize(f.count(col.foo), my_count=f.count(col.bar)).render(),
        )

    def test_summarize_by(self):
        self.assertEqual(
            " | summarize count(foo), my_count = count(bar) by bla, bin(date, 1), time_range = (bin(time, 10))",
            Query().summarize(f.count(col.foo), my_count=f.count(col.bar)).by(col.bla, f.bin(col.date, 1),
                                                                              time_range=f.bin(col.time, 10)).render(),
        )

    def test_summarize_by_expression(self):
        self.assertEqual(
            " | summarize count(foo) by tostring(asd)",
            Query().summarize(f.count(col.foo)).by(f.tostring(col.asd)).render(),
        )

    def test_mv_expand(self):
        self.assertEqual(
            " | mv-expand a, b, c",
            Query().mv_expand(col.a, col.b, col.c).render(),
        )

    def test_mv_expand_args(self):
        self.assertEqual(
            " | mv-expand bagexpansion=bag with_itemindex=foo a, b, c limit 4",
            Query().mv_expand(col.a, col.b, col.c, bag_expansion=BagExpansion.BAG, with_item_index=col.foo,
                              limit=4).render(),
        )

    def test_mv_expand_no_args(self):
        self.assertRaises(
            ValueError("Please specify one or more columns for mv-expand"),
            Query().mv_expand
        )

    def test_limit(self):
        self.assertEqual(
            " | limit 3",
            Query().limit(3).render(),
        )

    def test_sample(self):
        self.assertEqual(
            " | sample 3",
            Query().sample(3).render(),
        )

    def test_count(self):
        self.assertEqual(
            " | count",
            Query().count().render(),
        )

    def test_project(self):
        self.assertEqual(
            " | project a, bc",
            Query().project(col.a, col.bc).render(),
        )

    def test_project_with_expression(self):
        self.assertEqual(
            " | project foo = (bar * 4)",
            Query().project(foo=col.bar * 4).render(),
        )

    def test_project_assign_to_multiple_columns(self):
        self.assertEqual(
            " | project (foo, bar) = arr",
            Query().project(col.arr.assign_to(col.foo, col.bar)).render(),
        )

    def test_project_unspecified_column(self):
        self.assertEqual(
            " | project (foo + bar)",
            Query().project(col.foo + col.bar).render(),
        )

    def test_project_away(self):
        self.assertEqual(
            " | project-away a, bc",
            Query().project_away(col.a, col.bc).render(),
        )

    def test_project_away_wildcard(self):
        self.assertEqual(
            " | project-away a, b*",
            Query().project_away(col.a, "b*").render(),
        )

    def test_project_rename(self):
        self.assertEqual(
            " | project-rename a = b, c = d",
            Query().project_rename(a=col.b, c=col.d).render(),
        )

    def test_custom(self):
        self.assertEqual(
            " | some custom query",
            Query().custom("some custom query").render(),
        )

    def test_distinct(self):
        self.assertEqual(
            " | distinct a, b * 2",
            Query().distinct(col.a, col.b * 2).render(),
        )

    def test_distinct_all(self):
        self.assertEqual(
            " | distinct *",
            Query().distinct_all().render(),
        )

    def test_udf(self):
        # noinspection PyGlobalUndefined
        def func():
            global result
            global df

            result = df
            result['StateZone'] = result["State"] + result["Zone"]

        # TODO assert
        Query().evaluate(func, "typeof(*, StateZone: string)").render()



from typing import List, Tuple
from unittest.mock import patch
from urllib.parse import urljoin

from azure.kusto.data.request import KustoClient, ClientRequestProperties

from pykusto.expressions import column_generator as col
from pykusto.client import PyKustoClient
from pykusto.query import Query
from test.test_base import TestBase


# noinspection PyMissingConstructor
class MockKustoClient(KustoClient):
    executions: List[Tuple[str, str, ClientRequestProperties]]

    def __init__(self, cluster="https://test_cluster.kusto.windows.net"):
        self.executions = []
        self._query_endpoint = urljoin(cluster, "/v2/rest/query")

    def execute(self, database: str, rendered_query: str, properties: ClientRequestProperties = None):
        self.executions.append((database, rendered_query, properties))


class TestTable(TestBase):
    def test_single_table(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table']
        Query(table).take(5).execute()
        self.assertEqual(
            [('test_db', 'test_table | take 5', None)],
            mock_kusto_client.executions
        )

    def test_execute_no_table(self):
        self.assertRaises(
            RuntimeError("No table supplied"),
            Query().take(5).execute
        )

    def test_execute_already_bound(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table']

        self.assertRaises(
            RuntimeError("This table is already bound to a query"),
            Query(table).take(5).execute,
            table
        )

    def test_single_table_on_execute(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table']
        Query().take(5).execute(table)
        self.assertEqual(
            [('test_db', 'test_table | take 5', None)],
            mock_kusto_client.executions,
        )

    def test_union_table(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db'].get_tables('test_table1', 'test_table2')
        Query(table).take(5).execute()
        self.assertEqual(
            [('test_db', 'union test_table1, test_table2 | take 5', None)],
            mock_kusto_client.executions,
        )

    def test_union_table_with_wildcard(self):
        mock_kusto_client = MockKustoClient()
        table = PyKustoClient(mock_kusto_client)['test_db']['test_table_*']
        Query(table).take(5).execute()
        self.assertEqual(
            [('test_db', 'union test_table_* | take 5', None)],
            mock_kusto_client.executions,
        )

    def test_default_authentication(self):
        mock_kusto_client = MockKustoClient()
        with patch('pykusto.client.PyKustoClient._get_client_for_cluster', lambda s, cluster: mock_kusto_client):
            table = PyKustoClient('https://help.kusto.windows.net/')['test_db']['test_table']
            Query().take(5).execute(table)
        self.assertIs(
            mock_kusto_client,
            table.database.client._client,
        )
        self.assertEqual(
            [('test_db', 'test_table | take 5', None)],
            mock_kusto_client.executions,
        )

    def test_cross_cluster_join(self):
        mock_kusto_client_1 = MockKustoClient("https://one.kusto.windows.net")
        mock_kusto_client_2 = MockKustoClient("https://two.kusto.windows.net")

        table1 = PyKustoClient(mock_kusto_client_1)['test_db_1']['test_table_1']
        table2 = PyKustoClient(mock_kusto_client_2)['test_db_2']['test_table_2']
        Query(table1).take(5).join(Query(table2).take(6)).on(col.foo).execute()
        self.assertEqual(
            [('test_db_1', 'test_table_1 | take 5 | join  (cluster("two.kusto.windows.net").database("test_db_2").table("test_table_2") | take 6) on foo', None)],
            mock_kusto_client_1.executions,
        )

    def test_cross_cluster_join_with_union(self):
        mock_kusto_client_1 = MockKustoClient("https://one.kusto.windows.net")
        mock_kusto_client_2 = MockKustoClient("https://two.kusto.windows.net")

        table1 = PyKustoClient(mock_kusto_client_1)['test_db_1']['test_table_1']
        table2 = PyKustoClient(mock_kusto_client_2)['test_db_2'].get_tables('test_table_2_*')
        Query(table1).take(5).join(Query(table2).take(6)).on(col.foo).execute()
        self.assertEqual(
            [('test_db_1',
              'test_table_1 | take 5 | join  (union cluster("two.kusto.windows.net").database("test_db_2").table("test_table_2_*") | take 6) on foo',
              None)],
            mock_kusto_client_1.executions,
        )

    def test_cross_cluster_join_with_union_2(self):
        mock_kusto_client_1 = MockKustoClient("https://one.kusto.windows.net")
        mock_kusto_client_2 = MockKustoClient("https://two.kusto.windows.net")

        table1 = PyKustoClient(mock_kusto_client_1)['test_db_1']['test_table_1']
        table2 = PyKustoClient(mock_kusto_client_2)['test_db_2'].get_tables('test_table_2_*', 'test_table_3_*')
        Query(table1).take(5).join(Query(table2).take(6)).on(col.foo).execute()
        self.assertEqual(
            [('test_db_1',
              'test_table_1 | take 5 | join  (union cluster("two.kusto.windows.net").database("test_db_2").table("test_table_2_*"), cluster("two.kusto.windows.net").database("test_db_2").table("test_table_3_*") | take 6) on foo',
              None)],
            mock_kusto_client_1.executions,
        )



import logging
import sys
from typing import Callable
from unittest import TestCase

from pykusto.logger import logger


class TestBase(TestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(
            stream=sys.stdout,
            level=logging.DEBUG,
            format='%(asctime)s %(levelname)5s %(message)s'
        )

    def setUp(self):
        logger.info("Running test: " + self._testMethodName)

    def assertRaises(self, expected_exception: BaseException, test_callable: Callable, *args, **kwargs):
        """
        This method overrides the one in `unittest.case.TestCase`.

        Instead of providing it with an exception type, you provide it with an exception instance that contains
        the expected message.
        """
        expected_exception_type = type(expected_exception)
        expected_exception_message = str(expected_exception)
        with super().assertRaises(expected_exception_type) as cm:
            test_callable(*args, **kwargs)
        self.assertEqual(
            expected_exception_message,
            str(cm.exception)
        )



from setuptools import setup, find_packages

setup(
    name='pykusto',
    version='0.0.9',
    packages=find_packages(exclude=['test']),
    url='https://github.com/Azure/pykusto',
    license='MIT License',
    author='Microsoft Corporation',
    author_email='yomost@microsoft.com',
    description='Advanced python SDK for Azure Data Explorer',
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    keywords="kusto client library query",
    install_requires=['azure-kusto-data==0.0.36'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
    ],
)



import datetime

from pykusto import functions as f
from pykusto.expressions import column_generator as col
from pykusto.query import Query
from test.test_base import TestBase


# TODO dcount_hll

class TestFunction(TestBase):
    def test_acos(self):
        self.assertEqual(
            " | where (acos(foo)) > 4",
            Query().where(f.acos(col.foo) > 4).render()
        )

    def test_ago(self):
        self.assertEqual(
            " | where foo > (ago(time(4.0:0:0.0)))",
            Query().where(col.foo > f.ago(datetime.timedelta(4))).render()
        )
        self.assertEqual(
            " | where foo > (ago(timespan))",
            Query().where(col.foo > f.ago(col.timespan)).render()
        )

    def test_array_length(self):
        self.assertEqual(
            " | where (array_length(foo)) > 4",
            Query().where(f.array_length(col.foo) > 4).render()
        )

    def test_bag_keys(self):
        self.assertEqual(
            " | where (array_length(bag_keys(foo))) > 4",
            Query().where(f.bag_keys(col.foo).array_length() > 4).render()
        )

    def test_case(self):
        self.assertEqual(
            " | extend bucket = case(foo <= 3, \"Small\", foo <= 10, \"Medium\", \"Large\")",
            Query().extend(bucket=f.case(col.foo <= 3, "Small", col.foo <= 10, 'Medium', 'Large')).render()
        )

    def test_ceiling(self):
        self.assertEqual(
            " | where (ceiling(foo)) > 4",
            Query().where(f.ceiling(col.foo) > 4).render()
        )

    def test_cos(self):
        self.assertEqual(
            " | where (cos(foo)) > 4",
            Query().where(f.cos(col.foo) > 4).render()
        )

    def test_endofday(self):
        self.assertEqual(
            " | where (endofday(foo)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.endofday(col.foo) > datetime.datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (endofday(foo, 2)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.endofday(col.foo, 2) > datetime.datetime(2019, 7, 23)).render()
        )

    def test_endofmonth(self):
        self.assertEqual(
            " | where (endofmonth(foo)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.endofmonth(col.foo) > datetime.datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (endofmonth(foo, 2)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.endofmonth(col.foo, 2) > datetime.datetime(2019, 7, 23)).render()
        )

    def test_endofweek(self):
        self.assertEqual(
            " | where (endofweek(foo)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.endofweek(col.foo) > datetime.datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (endofweek(foo, 2)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.endofweek(col.foo, 2) > datetime.datetime(2019, 7, 23)).render()
        )

    def test_exp(self):
        self.assertEqual(
            " | where (exp(foo)) > 4",
            Query().where(f.exp(col.foo) > 4).render()
        )

    def test_exp10(self):
        self.assertEqual(
            " | where (exp10(foo)) > 4",
            Query().where(f.exp10(col.foo) > 4).render()
        )

    def test_exp2(self):
        self.assertEqual(
            " | where (exp2(foo)) > 4",
            Query().where(f.exp2(col.foo) > 4).render()
        )

    def test_floor(self):
        self.assertEqual(
            " | where (floor(foo, time(0.12:0:0.0))) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.floor(col.foo, datetime.timedelta(0.5)) > datetime.datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (floor(foo, 0.1)) > 3",
            Query().where(f.floor(col.foo, 0.1) > 3).render()
        )
        self.assertEqual(
            " | where (floor(foo, 0.1)) > 3",
            Query().where(col.foo.floor(0.1) > 3).render()
        )

    def test_format_datetime(self):
        self.assertEqual(
            " | where (format_datetime(foo, \"yy-MM-dd [HH:mm:ss]\")) == \"2019-07-23 00:00:00\"",
            Query().where(f.format_datetime(col.foo, 'yy-MM-dd [HH:mm:ss]') == '2019-07-23 00:00:00').render()
        )

    def test_format_timespan(self):
        self.assertEqual(
            " | where (format_timespan(foo, \"h:m:s.fffffff\")) == \"2:3:4.1234500\"",
            Query().where(f.format_timespan(col.foo, 'h:m:s.fffffff') == "2:3:4.1234500").render()
        )

    def test_getmonth(self):
        self.assertEqual(
            " | where (getmonth(foo)) > 3",
            Query().where(f.getmonth(col.foo) > 3).render()
        )

    def test_gettype(self):
        self.assertEqual(
            " | where (gettype(foo)) == \"datetime\"",
            Query().where(f.gettype(col.foo) == 'datetime').render()
        )

    def test_getyear(self):
        self.assertEqual(
            " | where (getyear(foo)) > 2019",
            Query().where(f.getyear(col.foo) > 2019).render()
        )

    def test_hourofday(self):
        self.assertEqual(
            " | where (hourofday(foo)) == 3",
            Query().where(f.hourofday(col.foo) == 3).render()
        )

    def test_hash(self):
        self.assertEqual(
            " | where (hash(foo)) == 3",
            Query().where(f.hash(col.foo) == 3).render()
        )

    def test_hash_sha256(self):
        self.assertEqual(
            " | where (hash_sha256(foo)) == 3",
            Query().where(f.hash_sha256(col.foo) == 3).render()
        )

    def test_isempty(self):
        self.assertEqual(
            " | where isempty(foo)",
            Query().where(f.isempty(col.foo)).render()
        )

    def test_isfinite(self):
        self.assertEqual(
            " | where isfinite(foo)",
            Query().where(f.isfinite(col.foo)).render()
        )

    def test_isinf(self):
        self.assertEqual(
            " | where isinf(foo)",
            Query().where(f.isinf(col.foo)).render()
        )

    def test_isnan(self):
        self.assertEqual(
            " | where isnan(foo)",
            Query().where(f.isnan(col.foo)).render()
        )

    def test_isnotempty(self):
        self.assertEqual(
            " | where isnotempty(foo)",
            Query().where(f.isnotempty(col.foo)).render()
        )

    def test_isnotnull(self):
        self.assertEqual(
            " | where isnotnull(foo)",
            Query().where(f.isnotnull(col.foo)).render()
        )

    def test_isnull(self):
        self.assertEqual(
            " | where isnull(foo)",
            Query().where(f.isnull(col.foo)).render()
        )

    def test_isutf8(self):
        self.assertEqual(
            " | where isutf8(foo)",
            Query().where(f.isutf8(col.foo)).render()
        )

    def test_log(self):
        self.assertEqual(
            " | where (log(foo)) < 3",
            Query().where(f.log(col.foo) < 3).render()
        )

    def test_log10(self):
        self.assertEqual(
            " | where (log10(foo)) < 3",
            Query().where(f.log10(col.foo) < 3).render(),
        )

    def test_log2(self):
        self.assertEqual(
            " | where (log2(foo)) < 3",
            Query().where(f.log2(col.foo) < 3).render()
        )

    def test_loggamma(self):
        self.assertEqual(
            " | where (loggamma(foo)) < 3",
            Query().where(f.loggamma(col.foo) < 3).render()
        )

    def test_make_datetime(self):
        self.assertEqual(
            " | where (make_datetime(y, m, d, 0, 0, 0)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.make_datetime(col.y, col.m, col.d) > datetime.datetime(2019, 7, 23)).render()
        )
        self.assertEqual(
            " | where (make_datetime(y, m, d, h, 0, 0)) > datetime(2019-07-23 05:00:00.000000)",
            Query().where(f.make_datetime(col.y, col.m, col.d, col.h) > datetime.datetime(2019, 7, 23, 5)).render()
        )
        self.assertEqual(
            " | where (make_datetime(y, m, d, h, min, 0)) > datetime(2019-07-23 05:29:00.000000)",
            Query().where(
                f.make_datetime(col.y, col.m, col.d, col.h, col.min) > datetime.datetime(2019, 7, 23, 5, 29)).render()
        )
        self.assertEqual(
            " | where (make_datetime(y, m, d, h, min, sec)) > datetime(2019-07-23 05:29:15.000000)",
            Query().where(
                f.make_datetime(col.y, col.m, col.d, col.h, col.min, col.sec) > datetime.datetime(2019, 7, 23, 5, 29,
                                                                                                  15)).render()
        )

    def test_now(self):
        self.assertEqual(
            " | where foo < (now())",
            Query().where(col.foo < f.now()).render()
        )
        self.assertEqual(
            " | where foo < (now(time(-3.0:0:0.0)))",
            Query().where(col.foo < f.now(datetime.timedelta(-3))).render()
        )

    def test_parse_json_to_string(self):
        self.assertEqual(
            ' | where (tostring(parse_json(foo))) contains "ABC"',
            Query().where(f.parse_json(col.foo).to_string().contains('ABC')).render()
        )

    def test_parse_json_brackets(self):
        self.assertEqual(
            ' | where (tostring(parse_json(foo)["bar"])) contains "ABC"',
            Query().where(f.parse_json(col.foo)['bar'].to_string().contains('ABC')).render()
        )

    def test_parse_json_dot(self):
        self.assertEqual(
            ' | where (tostring(parse_json(foo).bar)) contains "ABC"',
            Query().where(f.parse_json(col.foo).bar.to_string().contains('ABC')).render()
        )

    def test_parse_json_number_expression(self):
        self.assertEqual(
            ' | where (todouble(parse_json(foo).bar)) > 4',
            Query().where(f.todouble(f.parse_json(col.foo).bar) > 4).render()
        )

    def test_parse_json_array(self):
        self.assertEqual(
            ' | where (parse_json(foo)[2]) == 3',
            Query().where(f.parse_json(col.foo)[2] == 3).render()
        )

    def test_parse_json_nesting(self):
        self.assertEqual(
            ' | where (parse_json(foo)["a"].b[2]) contains "bar"',
            Query().where(f.parse_json(col.foo)['a'].b[2].contains('bar')).render())

    def test_pow(self):
        self.assertEqual(
            " | where (pow(foo, bar)) > 3",
            Query().where(f.pow(col.foo, col.bar) > 3).render()
        )

    def test_round(self):
        self.assertEqual(
            " | where (round(foo, bar)) == 3",
            Query().where(f.round(col.foo, col.bar) == 3).render()
        )

    def test_sign(self):
        self.assertEqual(
            " | where (sign(foo)) == 1",
            Query().where(f.sign(col.foo) == 1).render()
        )

    def test_sqrt(self):
        self.assertEqual(
            " | where (sqrt(foo)) > 1",
            Query().where(f.sqrt(col.foo) > 1).render()
        )

    def test_startofday(self):
        self.assertEqual(
            " | where (startofday(foo)) > datetime(2019-07-23 00:00:00.000000)",
            Query().where(f.startofday(col.foo) > datetime.datetime(2019, 7, 23)).render()
        )

    def test_startofmonth(self):
        self.assertEqual(
            " | where (startofmonth(foo)) > datetime(2019-07-01 00:00:00.000000)",
            Query().where(f.startofmonth(col.foo) > datetime.datetime(2019, 7, 1)).render()
        )

    def test_startofweek(self):
        self.assertEqual(
            " | where (startofweek(foo)) > datetime(2019-07-08 00:00:00.000000)",
            Query().where(f.startofweek(col.foo) > datetime.datetime(2019, 7, 8)).render()
        )

    def test_startofyear(self):
        self.assertEqual(
            " | where (startofyear(foo)) > datetime(2019-01-01 00:00:00.000000)",
            Query().where(f.startofyear(col.foo) > datetime.datetime(2019, 1, 1)).render()
        )

    def test_strcat(self):
        self.assertEqual(
            " | extend (strcat(\"hello\", \",\", foo, \"!\"))",
            Query().extend(f.strcat("hello", ',', col.foo, '!')).render()
        )
        self.assertEqual(
            " | extend (strcat(foo, \"!\"))",
            Query().extend(f.strcat(col.foo, '!')).render()
        )

    def test_strcat_delim(self):
        self.assertEqual(
            " | extend (strcat_delim(\"-\", \"hello\", \",\", foo, \"!\"))",
            Query().extend(f.strcat_delim('-', "hello", ',', col.foo, '!')).render()
        )
        self.assertEqual(
            " | extend (strcat_delim(\"-\", \",\", foo))",
            Query().extend(f.strcat_delim('-', ',', col.foo)).render()
        )

    def test_strcat_array(self):
        self.assertEqual(
            " | where (strcat_array(foo, \",\")) == \"A,B,C\"",
            Query().where(f.strcat_array(col.foo, ',') == 'A,B,C').render()
        )
        self.assertEqual(
            " | where (strcat_array(dynamic([\"A\", \"B\", \"C\"]), \",\")) == \"A,B,C\"",
            Query().where(f.strcat_array(['A', 'B', 'C'], ',') == 'A,B,C').render()
        )

    def test_strcmp(self):
        self.assertEqual(
            " | where (strcmp(foo, bar)) == 1",
            Query().where(f.strcmp(col.foo, col.bar) == 1).render()
        )

    def test_string_size(self):
        self.assertEqual(
            " | where (string_size(foo)) == 1",
            Query().where(f.string_size(col.foo) == 1).render()
        )

    def test_strlen(self):
        self.assertEqual(
            " | where (strlen(foo)) == 1",
            Query().where(f.strlen(col.foo) == 1).render()
        )

    def test_strrep(self):
        self.assertEqual(
            " | where (strrep(foo, bar)) == \"ABCABC\"",
            Query().where(f.strrep(col.foo, col.bar) == 'ABCABC').render()
        )
        self.assertEqual(
            " | where (strrep(foo, bar, \",\")) == \"ABC,ABC\"",
            Query().where(f.strrep(col.foo, col.bar, ',') == 'ABC,ABC').render()
        )
        self.assertEqual(
            " | where (strrep(foo, bar, fam)) == \"ABC,ABC\"",
            Query().where(f.strrep(col.foo, col.bar, col.fam) == 'ABC,ABC').render()
        )

    def test_substring(self):
        self.assertEqual(
            " | where (substring(foo, bar)) == \"ABCABC\"",
            Query().where(f.substring(col.foo, col.bar) == 'ABCABC').render()
        )
        self.assertEqual(
            " | where (substring(foo, bar, 4)) == \"ABC,ABC\"",
            Query().where(f.substring(col.foo, col.bar, 4) == 'ABC,ABC').render()
        )

    def test_split(self):
        self.assertEqual(
            ' | extend foo = (split(bar, "_", 3))',
            Query().extend(foo=f.split(col.bar, "_", 3)).render()
        )
        self.assertEqual(
            ' | extend foo = (split(bar, "_")[3])',
            Query().extend(foo=f.split(col.bar, "_")[3]).render()
        )
        self.assertEqual(
            ' | extend foo = (split("1_2", "_")[3])',
            Query().extend(foo=f.split("1_2", "_")[3]).render()
        )

    def test_tobool(self):
        self.assertEqual(
            " | where tobool(foo)",
            Query().where(f.tobool(col.foo)).render()
        )

    def test_toboolean(self):
        self.assertEqual(
            " | where toboolean(foo)",
            Query().where(f.toboolean(col.foo)).render()
        )

    # def test_todatetime(self):
    #     self.assertEqual(
    #         Query().where(f.todatetime(col.foo) > datetime.datetime(2019, 7, 23)).render(),
    #         " | where (startofday(foo)) > datetime(2019-07-23 00:00:00.000000)")
    #     self.assertEqual(
    #         Query().where(f.todatetime('') > datetime.datetime(2019, 7, 23)).render(),
    #         " | where (startofday(foo)) > datetime(2019-07-23 00:00:00.000000)")
    # def todatetime(expr: StringType) -> DatetimeExpression:
    #     return DatetimeExpression(KQL('todatetime({})'.format(_subexpr_to_kql(expr))))

    def test_todouble(self):
        self.assertEqual(
            " | where (todouble(foo)) > 0.2",
            Query().where(f.todouble(col.foo) > 0.2).render()
        )

    # ------------------------------------------------------
    # Aggregative Functions
    # ------------------------------------------------------

    def test_any(self):
        self.assertEqual(
            " | summarize any(foo, bar, fam)",
            Query().summarize(f.any(col.foo, col.bar, col.fam)).render()
        )

    def test_arg_max(self):
        self.assertEqual(
            " | summarize arg_max(foo, bar, fam)",
            Query().summarize(f.arg_max(col.foo, col.bar, col.fam)).render()
        )

    def test_arg_min(self):
        self.assertEqual(
            " | summarize arg_min(foo, bar, fam)",
            Query().summarize(f.arg_min(col.foo, col.bar, col.fam)).render()
        )

    def test_avg(self):
        self.assertEqual(
            " | summarize avg(foo)",
            Query().summarize(f.avg(col.foo)).render()
        )
        self.assertEqual(
            " | summarize avg(foo) - 5",
            Query().summarize(f.avg(col.foo) - 5).render()
        )

    def test_avgif(self):
        self.assertEqual(
            " | summarize avgif(foo, bar)",
            Query().summarize(f.avgif(col.foo, col.bar)).render()
        )

    def test_bin(self):
        self.assertEqual(
            " | summarize avg(foo) by bin(bar, 0.1)",
            Query().summarize(f.avg(col.foo)).by(f.bin(col.bar, 0.1)).render()
        )
        self.assertEqual(
            " | summarize avg(foo) by bin(bar, time(0.12:0:0.0))",
            Query().summarize(f.avg(col.foo)).by(f.bin(col.bar, datetime.timedelta(0.5))).render()
        )

    def test_bin_at(self):
        self.assertEqual(
            " | summarize avg(foo) by bin_at(bar, 0.1, 1)",
            Query().summarize(f.avg(col.foo)).by(f.bin_at(col.bar, 0.1, 1)).render()
        )
        self.assertEqual(
            " | summarize avg(foo) by bin_at(bar, time(0.12:0:0.0), time(0.2:24:0.0))",
            Query().summarize(f.avg(col.foo)).by(f.bin_at(col.bar,
                                                          datetime.timedelta(0.5),
                                                          datetime.timedelta(0.1))).render()
        )
        self.assertEqual(
            " | summarize avg(foo) by bin_at(bar, time(0.12:0:0.0), datetime(2019-07-08 00:00:00.000000))",
            Query().summarize(f.avg(col.foo)).by(f.bin_at(col.bar,
                                                          datetime.timedelta(0.5),
                                                          datetime.datetime(2019, 7, 8))).render()
        )

    def test_bin_auto(self):
        self.assertEqual(
            " | summarize avg(foo) by bin_auto(bar)",
            Query().summarize(f.avg(col.foo)).by(f.bin_auto(col.bar)).render()
        )

    def test_count(self):
        self.assertEqual(
            " | summarize count()",
            Query().summarize(f.count()).render()
        )
        self.assertEqual(
            " | summarize count(foo)",
            Query().summarize(f.count(col.foo)).render()
        )

    def test_countif(self):
        self.assertEqual(
            " | summarize countif(foo == 1)",
            Query().summarize(f.countif(col.foo == 1)).render()
        )

    def test_dcount(self):
        self.assertEqual(
            " | summarize dcount(foo)",
            Query().summarize(f.dcount(col.foo)).render()
        )
        acc = 0.1
        self.assertEqual(
            " | summarize dcount(foo, 0.1)",
            Query().summarize(f.dcount(col.foo, acc)).render()
        )

    # def test_hll(self):
    #     self.assertEqual(
    #         " | summarize hll(foo)",
    #         Query().summarize(f.hll(col.foo)).render()
    #         )
    #     acc = 0.1
    #     self.assertEqual(
    #         " | summarize hll(foo, 0.1)",
    #         Query().summarize(f.hll(col.foo, acc)).render()
    #         )
    #
    # def test_hll_merge(self):
    #     self.assertEqual(
    #         " | summarize hll_merge(foo)",
    #         Query().summarize(f.hll_merge(col.foo)).render()
    #         )

    def test_make_bag(self):
        self.assertEqual(
            " | summarize make_bag(foo)",
            Query().summarize(f.make_bag(col.foo)).render()
        )

    def test_make_list(self):
        self.assertEqual(
            " | summarize make_list(foo)",
            Query().summarize(f.make_list(col.foo)).render()
        )

    def test_make_set(self):
        self.assertEqual(
            " | summarize make_set(foo)",
            Query().summarize(f.make_set(col.foo)).render()
        )

    def test_max(self):
        self.assertEqual(
            " | summarize max(foo)",
            Query().summarize(f.max(col.foo)).render()
        )

    def test_min(self):
        self.assertEqual(
            " | summarize min(foo)",
            Query().summarize(f.min(col.foo)).render()
        )

    def test_percentile(self):
        self.assertEqual(
            " | summarize percentiles(foo, 5)",
            Query().summarize(f.percentiles(col.foo, 5)).render()
        )

    def test_percentiles(self):
        self.assertEqual(
            " | summarize percentiles(foo, 5, 50, 95)",
            Query().summarize(f.percentiles(col.foo, 5, 50, 95)).render()
        )

    def test_stdev(self):
        self.assertEqual(
            " | summarize stdev(foo)",
            Query().summarize(f.stdev(col.foo)).render()
        )

    def test_stdevif(self):
        self.assertEqual(
            " | summarize stdevif(foo, bar)",
            Query().summarize(f.stdevif(col.foo, col.bar)).render()
        )

    def test_stdevp(self):
        self.assertEqual(
            " | summarize stdevp(foo)",
            Query().summarize(f.stdevp(col.foo)).render()
        )

    def test_sum(self):
        self.assertEqual(
            " | summarize sum(foo)",
            Query().summarize(f.sum(col.foo)).render()
        )

    def test_sumif(self):
        self.assertEqual(
            " | summarize sumif(foo, bar)",
            Query().summarize(f.sumif(col.foo, col.bar)).render()
        )

    def test_variance(self):
        self.assertEqual(
            " | summarize variance(foo)",
            Query().summarize(f.variance(col.foo)).render()
        )

    def test_varianceif(self):
        self.assertEqual(
            " | summarize varianceif(foo, bar)",
            Query().summarize(f.varianceif(col.foo, col.bar)).render()
        )

    def test_variancep(self):
        self.assertEqual(
            " | summarize variancep(foo)",
            Query().summarize(f.variancep(col.foo)).render()
        )

    def test_nesting(self):
        self.assertEqual(
            " | summarize active_days = dcount(bin(timestamp, time(1.0:0:0.0)))",
            Query().summarize(active_days=f.dcount(f.bin(col.timestamp, datetime.timedelta(1)))).render()
        )

    def test_iff(self):
        self.assertEqual(
            " | project foo = (iff(foo > (ago(time(2.0:0:0.0))), time(3.0:0:0.0), time(4.0:0:0.0)))",
            Query().project(foo=f.iff(col.foo > f.ago(datetime.timedelta(2)), datetime.timedelta(3), datetime.timedelta(4))).render()
        )

    def test_iif(self):
        # iif is just an alias to iff
        self.assertEqual(
            " | project foo = (iff(foo > (ago(time(2.0:0:0.0))), time(3.0:0:0.0), time(4.0:0:0.0)))",
            Query().project(foo=f.iif(col.foo > f.ago(datetime.timedelta(2)), datetime.timedelta(3), datetime.timedelta(4))).render()
        )
